'''Splitter
======

.. versionadded:: 1.5.0

.. image:: images/splitter.jpg
    :align: right

The :class:`Splitter` is a widget that helps you re-size its child
widget/layout by letting you re-size it via dragging the boundary or
double tapping the boundary. This widget is similar to the
:class:`~kivy.uix.scrollview.ScrollView` in that it allows only one
child widget.

Usage::

    splitter = Splitter(sizable_from = 'right')
    splitter.add_widget(layout_or_widget_instance)
    splitter.min_size = 100
    splitter.max_size = 250

To change the size of the strip/border used for resizing::

    splitter.strip_size = '10pt'

To change its appearance::

    splitter.strip_cls = your_custom_class

You can also change the appearance of the `strip_cls`, which defaults to
:class:`SplitterStrip`, by overriding the `kv` rule in your app:

.. code-block:: kv

    <SplitterStrip>:
        horizontal: True if self.parent and self.parent.sizable_from[0] \
in ('t', 'b') else False
        background_normal: 'path to normal horizontal image' \
if self.horizontal else 'path to vertical normal image'
        background_down: 'path to pressed horizontal image' \
if self.horizontal else 'path to vertical pressed image'

'''


__all__ = ('Splitter', )

from kivy.compat import string_types
from kivy.factory import Factory
from kivy.uix.button import Button
from kivy.properties import (OptionProperty, NumericProperty, ObjectProperty,
                             ListProperty, BooleanProperty)
from kivy.uix.boxlayout import BoxLayout


class SplitterStrip(Button):
    '''Class used for tbe graphical representation of a
    :class:`kivy.uix.splitter.SplitterStripe`.
    '''
    pass


class Splitter(BoxLayout):
    '''See module documentation.

    :Events:
        `on_press`:
            Fired when the splitter is pressed.
        `on_release`:
            Fired when the splitter is released.

    .. versionchanged:: 1.6.0
        Added `on_press` and `on_release` events.

    '''

    border = ListProperty([4, 4, 4, 4])
    '''Border used for the
    :class:`~kivy.graphics.vertex_instructions.BorderImage`
    graphics instruction.

    This must be a list of four values: (bottom, right, top, left).
    Read the BorderImage instructions for more information about how
    to use it.

    :attr:`border` is a :class:`~kivy.properties.ListProperty` and
    defaults to (4, 4, 4, 4).
    '''

    strip_cls = ObjectProperty(SplitterStrip)
    '''Specifies the class of the resize Strip.

    :attr:`strip_cls` is an :class:`kivy.properties.ObjectProperty` and
    defaults to :class:`~kivy.uix.splitter.SplitterStrip`, which is of type
    :class:`~kivy.uix.button.Button`.

    .. versionchanged:: 1.8.0
        If you set a string, the :class:`~kivy.factory.Factory` will be used to
        resolve the class.

    '''

    sizable_from = OptionProperty('left', options=(
        'left', 'right', 'top', 'bottom'))
    '''Specifies whether the widget is resizable. Options are:
    `left`, `right`, `top` or `bottom`

    :attr:`sizable_from` is an :class:`~kivy.properties.OptionProperty`
    and defaults to `left`.
    '''

    strip_size = NumericProperty('10pt')
    '''Specifies the size of resize strip

    :attr:`strp_size` is a :class:`~kivy.properties.NumericProperty`
    defaults to `10pt`
    '''

    min_size = NumericProperty('100pt')
    '''Specifies the minimum size beyond which the widget is not resizable.

    :attr:`min_size` is a :class:`~kivy.properties.NumericProperty` and
    defaults to `100pt`.
    '''

    max_size = NumericProperty('500pt')
    '''Specifies the maximum size beyond which the widget is not resizable.

    :attr:`max_size` is a :class:`~kivy.properties.NumericProperty`
    and defaults to `500pt`.
    '''

    _parent_proportion = NumericProperty(0.)
    '''(internal) Specifies the distance that the slider has travelled
    across its parent, used to automatically maintain a sensible
    position if the parent is resized.

    :attr:`_parent_proportion` is a
    :class:`~kivy.properties.NumericProperty` and defaults to 0.

    .. versionadded:: 1.9.0
    '''

    _bound_parent = ObjectProperty(None, allownone=True)
    '''(internal) References the widget whose size is currently being
    tracked by :attr:`_parent_proportion`.

    :attr:`_bound_parent` is a
    :class:`~kivy.properties.ObjectProperty` and defaults to None.

    .. versionadded:: 1.9.0
    '''

    keep_within_parent = BooleanProperty(False)
    '''If True, will limit the splitter to stay within its parent widget.

    :attr:`keep_within_parent` is a
    :class:`~kivy.properties.BooleanProperty` and defaults to False.

    .. versionadded:: 1.9.0
    '''

    rescale_with_parent = BooleanProperty(False)
    '''If True, will automatically change size to take up the same
    proportion of the parent widget when it is resized, while
    staying within :attr:`min_size` and :attr:`max_size`. As long as
    these attributes can be satisfied, this stops the
    :class:`Splitter` from exceeding the parent size during rescaling.

    :attr:`rescale_with_parent` is a
    :class:`~kivy.properties.BooleanProperty` and defaults to False.

    .. versionadded:: 1.9.0
    '''

    __events__ = ('on_press', 'on_release')

    def __init__(self, **kwargs):
        self._container = None
        self._strip = None
        super(Splitter, self).__init__(**kwargs)

        do_size = self._do_size
        fbind = self.fbind
        fbind('max_size', do_size)
        fbind('min_size', do_size)
        fbind('parent', self._rebind_parent)

    def on_sizable_from(self, instance, sizable_from):
        if not instance._container:
            return

        sup = super(Splitter, instance)
        _strp = instance._strip
        if _strp:
            # remove any previous binds
            _strp.unbind(on_touch_down=instance.strip_down)
            _strp.unbind(on_touch_move=instance.strip_move)
            _strp.unbind(on_touch_up=instance.strip_up)
            self.unbind(disabled=_strp.setter('disabled'))

            sup.remove_widget(instance._strip)
        else:
            cls = instance.strip_cls
            if isinstance(cls, string_types):
                cls = Factory.get(cls)
            instance._strip = _strp = cls()

        sz_frm = instance.sizable_from[0]
        if sz_frm in ('l', 'r'):
            _strp.size_hint = None, 1
            _strp.width = instance.strip_size
            instance.orientation = 'horizontal'
            instance.unbind(strip_size=_strp.setter('width'))
            instance.bind(strip_size=_strp.setter('width'))
        else:
            _strp.size_hint = 1, None
            _strp.height = instance.strip_size
            instance.orientation = 'vertical'
            instance.unbind(strip_size=_strp.setter('height'))
            instance.bind(strip_size=_strp.setter('height'))

        index = 1
        if sz_frm in ('r', 'b'):
            index = 0
        sup.add_widget(_strp, index)

        _strp.bind(on_touch_down=instance.strip_down)
        _strp.bind(on_touch_move=instance.strip_move)
        _strp.bind(on_touch_up=instance.strip_up)
        _strp.disabled = self.disabled
        self.bind(disabled=_strp.setter('disabled'))

    def add_widget(self, widget, index=0):
        if self._container or not widget:
            return Exception('Splitter accepts only one Child')
        self._container = widget
        sz_frm = self.sizable_from[0]
        if sz_frm in ('l', 'r'):
            widget.size_hint_x = 1
        else:
            widget.size_hint_y = 1

        index = 0
        if sz_frm in ('r', 'b'):
            index = 1
        super(Splitter, self).add_widget(widget, index)
        self.on_sizable_from(self, self.sizable_from)

    def remove_widget(self, widget, *largs):
        super(Splitter, self).remove_widget(widget)
        if widget == self._container:
            self._container = None

    def clear_widgets(self):
        self.remove_widget(self._container)

    def strip_down(self, instance, touch):
        if not instance.collide_point(*touch.pos):
            return False
        touch.grab(self)
        self.dispatch('on_press')

    def on_press(self):
        pass

    def _rebind_parent(self, instance, new_parent):
        if self._bound_parent is not None:
            self._bound_parent.unbind(size=self.rescale_parent_proportion)
        if self.parent is not None:
            new_parent.bind(size=self.rescale_parent_proportion)
        self._bound_parent = new_parent
        self.rescale_parent_proportion()

    def rescale_parent_proportion(self, *args):
        if not self.parent:
            return
        if self.rescale_with_parent:
            parent_proportion = self._parent_proportion
            if self.sizable_from in ('top', 'bottom'):
                new_height = parent_proportion * self.parent.height
                self.height = max(self.min_size,
                                 min(new_height, self.max_size))
            else:
                new_width = parent_proportion * self.parent.width
                self.width = max(self.min_size, min(new_width, self.max_size))

    def _do_size(self, instance, value):
        if self.sizable_from[0] in ('l', 'r'):
            self.width = max(self.min_size, min(self.width, self.max_size))
        else:
            self.height = max(self.min_size, min(self.height, self.max_size))

    @staticmethod
    def _is_moving(sz_frm, diff, pos, minpos, maxpos):
        if sz_frm in ('l', 'b'):
            cmp = minpos
        else:
            cmp = maxpos
        if diff == 0:
            return False
        elif diff > 0 and pos <= cmp:
            return False
        elif diff < 0 and pos >= cmp:
            return False
        return True

    def strip_move(self, instance, touch):
        if touch.grab_current is not instance:
            return False
        max_size = self.max_size
        min_size = self.min_size
        sz_frm = self.sizable_from[0]

        if sz_frm in ('t', 'b'):
            diff_y = (touch.dy)
            self_y = self.y
            self_top = self.top
            if not self._is_moving(sz_frm, diff_y, touch.y, self_y, self_top):
                return
            if self.keep_within_parent:
                if sz_frm == 't' and (self_top + diff_y) > self.parent.top:
                    diff_y = self.parent.top - self_top
                elif sz_frm == 'b' and (self_y + diff_y) < self.parent.y:
                    diff_y = self.parent.y - self_y
            if sz_frm == 'b':
                diff_y *= -1
            if self.size_hint_y:
                self.size_hint_y = None
            if self.height > 0:
                self.height += diff_y
            else:
                self.height = 1

            height = self.height
            self.height = max(min_size, min(height, max_size))

            self._parent_proportion = self.height / self.parent.height
        else:
            diff_x = (touch.dx)
            self_x = self.x
            self_right = self.right
            if not self._is_moving(sz_frm, diff_x, touch.x, self_x, self_right):
                return
            if self.keep_within_parent:
                if sz_frm == 'l' and (self_x + diff_x) < self.parent.x:
                    diff_x = self.parent.x - self_x
                elif (sz_frm == 'r' and
                      (self_right + diff_x) > self.parent.right):
                    diff_x = self.parent.right - self_right
            if sz_frm == 'l':
                diff_x *= -1
            if self.size_hint_x:
                self.size_hint_x = None
            if self.width > 0:
                self.width += diff_x
            else:
                self.width = 1

            width = self.width
            self.width = max(min_size, min(width, max_size))

            self._parent_proportion = self.width / self.parent.width

    def strip_up(self, instance, touch):
        if touch.grab_current is not instance:
            return

        if touch.is_double_tap:
            max_size = self.max_size
            min_size = self.min_size
            sz_frm = self.sizable_from[0]
            s = self.size

            if sz_frm in ('t', 'b'):
                if self.size_hint_y:
                    self.size_hint_y = None
                if s[1] - min_size <= max_size - s[1]:
                    self.height = max_size
                else:
                    self.height = min_size
            else:
                if self.size_hint_x:
                    self.size_hint_x = None
                if s[0] - min_size <= max_size - s[0]:
                    self.width = max_size
                else:
                    self.width = min_size
        touch.ungrab(instance)
        self.dispatch('on_release')

    def on_release(self):
        pass


if __name__ == '__main__':
    from kivy.app import App
    from kivy.uix.button import Button
    from kivy.uix.floatlayout import FloatLayout

    class SplitterApp(App):

        def build(self):
            root = FloatLayout()
            bx = BoxLayout()
            bx.add_widget(Button())
            bx.add_widget(Button())
            bx2 = BoxLayout()
            bx2.add_widget(Button())
            bx2.add_widget(Button())
            bx2.add_widget(Button())
            spl = Splitter(
                size_hint=(1, .25),
                pos_hint={'top': 1},
                sizable_from='bottom')
            spl1 = Splitter(
                sizable_from='left',
                size_hint=(None, 1), width=90)
            spl1.add_widget(Button())
            bx.add_widget(spl1)
            spl.add_widget(bx)

            spl2 = Splitter(size_hint=(.25, 1))
            spl2.add_widget(bx2)
            spl2.sizable_from = 'right'
            root.add_widget(spl)
            root.add_widget(spl2)
            return root

    SplitterApp().run()



"""
AudioAndroid: Kivy audio implementation for Android using native API
"""

__all__ = ("SoundAndroidPlayer", )

from jnius import autoclass
from android import api_version
from kivy.core.audio import Sound, SoundLoader


MediaPlayer = autoclass("android.media.MediaPlayer")
AudioManager = autoclass("android.media.AudioManager")
if api_version >= 21:
    AudioAttributesBuilder = autoclass("android.media.AudioAttributes$Builder")


class SoundAndroidPlayer(Sound):
    @staticmethod
    def extensions():
        return ("mp3", "mp4", "aac", "3gp", "flac", "mkv", "wav", "ogg", "m4a",
                "gsm", "mid", "xmf", "mxmf", "rtttl", "rtx", "ota", "imy")

    def __init__(self, **kwargs):
        self._mediaplayer = None
        super(SoundAndroidPlayer, self).__init__(**kwargs)

    def load(self):
        self.unload()
        self._mediaplayer = MediaPlayer()
        if api_version >= 21:
            self._mediaplayer.setAudioAttributes(
                AudioAttributesBuilder()
                .setLegacyStreamType(AudioManager.STREAM_MUSIC)
                .build())
        else:
            self._mediaplayer.setAudioStreamType(AudioManager.STREAM_MUSIC)
        self._mediaplayer.setDataSource(self.source)
        self._mediaplayer.prepare()

    def unload(self):
        if self._mediaplayer:
            self._mediaplayer.release()
            self._mediaplayer = None

    def play(self):
        if not self._mediaplayer:
            return
        self._mediaplayer.start()
        super(SoundAndroidPlayer, self).play()

    def stop(self):
        if not self._mediaplayer:
            return
        self._mediaplayer.stop()
        self._mediaplayer.prepare()

    def seek(self, position):
        if not self._mediaplayer:
            return
        self._mediaplayer.seekTo(float(position) * 1000)

    def get_pos(self):
        if self._mediaplayer:
            return self._mediaplayer.getCurrentPosition() / 1000.
        return super(SoundAndroidPlayer, self).get_pos()

    def on_volume(self, instance, volume):
        if self._mediaplayer:
            volume = float(volume)
            self._mediaplayer.setVolume(volume, volume)

    def _get_length(self):
        if self._mediaplayer:
            return self._mediaplayer.getDuration() / 1000.
        return super(SoundAndroidPlayer, self)._get_length()

    def on_loop(self, instance, loop):
        if self._mediaplayer:
            self._mediaplayer.setLooping(loop)


SoundLoader.register(SoundAndroidPlayer)



#!/usr/bin/env python
# flake8: noqa
import os
import re
import random
from gimpfu import *

# Test suite configuration - key is test name, values are:
#
# alpha....: global alpha, used for all pixels except 't'
# patterns.: allowed v0 pattern characters (+ force include and exclude)
# *_IMAGE..: gimp layer types to export for this test, w/target extensions
TESTSUITE_CONFIG = {
    'OPAQUE': {
        'alpha': [255],
        'patterns': ('wxrgbcyp', None, None),
        RGB_IMAGE: ['xcf', 'png', 'tga', 'tiff', 'ppm',
                    'sgi', 'pcx', 'fits', 'ras'],
        INDEXED_IMAGE: ['xcf', 'png', 'tga', 'tiff', 'ppm', 'gif', 'ras'],
    },

    'GRAY-OPAQUE': {
        'alpha': [255],
        'patterns': ('0123456789ABCDEF', None, None),
        GRAY_IMAGE: ['xcf', 'png', 'tga', 'tiff',
                     'pgm', 'sgi', 'fits', 'ras'],
        INDEXED_IMAGE: ['xcf', 'png', 'tga', 'tiff', 'pgm', 'fits', 'ras'],
    },

    'BINARY': {
        'alpha': [255],
        'patterns': ('twrgbcyp', 't', None),
        RGBA_IMAGE: ['xcf', 'png', 'tga', 'ico', 'sgi'],
        INDEXEDA_IMAGE: ['xcf', 'png', 'tga', 'gif'],
    },
    'GRAY-BINARY': {
        'alpha': [255],
        'patterns': ('t123456789ABCDEF', 't', None),
        GRAYA_IMAGE: ['xcf', 'tga', 'png', 'sgi'],
        INDEXEDA_IMAGE: ['xcf', 'tga', 'png'],
    },

    "ALPHA": {
        'alpha': [0x7F, 0xF0],
        'patterns': ('twxrgbcyp', None, None),
        RGBA_IMAGE: ['xcf', 'png', 'tga', 'sgi'],
    },
    'GRAY-ALPHA': {
        'alpha': [0x7F, 0xF0],
        'patterns': ('t0123456789ABCDEF', None, None),
        GRAYA_IMAGE: ['xcf', 'png', 'tga', 'sgi'],
    },
}


# kivy image test protocol v0 data: key is pattern char, value is pixel w/o a
v0_PIXELS = {
    'w': [0xFF, 0xFF, 0xFF], 'x': [0x00, 0x00, 0x00], 'r': [0xFF, 0x00, 0x00],
    'g': [0x00, 0xFF, 0x00], 'b': [0x00, 0x00, 0xFF], 'y': [0xFF, 0xFF, 0x00],
    'c': [0x00, 0xFF, 0xFF], 'p': [0xFF, 0x00, 0xFF], '0': [0x00, 0x00, 0x00],
    '1': [0x11, 0x11, 0x11], '2': [0x22, 0x22, 0x22], '3': [0x33, 0x33, 0x33],
    '4': [0x44, 0x44, 0x44], '5': [0x55, 0x55, 0x55], '6': [0x66, 0x66, 0x66],
    '7': [0x77, 0x77, 0x77], '8': [0x88, 0x88, 0x88], '9': [0x99, 0x99, 0x99],
    'A': [0xAA, 0xAA, 0xAA], 'B': [0xBB, 0xBB, 0xBB], 'C': [0xCC, 0xCC, 0xCC],
    'D': [0xDD, 0xDD, 0xDD], 'E': [0xEE, 0xEE, 0xEE], 'F': [0xFF, 0xFF, 0xFF]}


# kivy image test protocol v0: return pixel data for given pattern char
def v0_pattern_pixel(char, alpha, fmt):
    if fmt == 'rgba':
        if char == 't':
            return [0, 0, 0, 0]
        return v0_PIXELS[char] + [alpha]
    if fmt == 'rgb':
        if char == 't':
            return [0, 0, 0]
        return v0_PIXELS[char]
    if fmt == 'gray':
        assert char in '0123456789ABCDEF'
        return [v0_PIXELS[char][0]]
    if fmt == 'graya':
        assert char in 't0123456789ABCDEF'
        if char == 't':
            return [0, 0]
        return [v0_PIXELS[char][0]] + [alpha]
    raise Exception('v0_pattern_pixel: unknown format {}'.format(fmt))


# kivy image test protocol v0: filename
def v0_filename(w, h, pat, alpha, fmtinfo, testname, ext):
    return 'v0_{}x{}_{}_{:02X}_{}_{}_gimp.{}'.format(
            w, h, pat, alpha, fmtinfo, testname, ext)


# Saves an image to one or more files. We can't specify these details when
# saving by extension. (This declaration is PEP8 compliant, 's all good)
def save_image(dirname, img, lyr, w, h, pat, alpha, v0_fmtinfo, testname, ext):
    def filename(fmtinfo_in=None):
        fmtinfo = fmtinfo_in and v0_fmtinfo + '-' + fmtinfo_in or v0_fmtinfo
        return v0_filename(w, h, pat, alpha, fmtinfo, testname, ext)

    def savepath(fn):
        return os.path.join(dirname, fn)

    if ext in ('ppm', 'pgm', 'pbm', 'pnm', 'pam'):
        fn = filename('ASCII')
        pdb.file_pnm_save(img, lyr, savepath(fn), fn, 0)
        fn = filename('RAW')
        pdb.file_pnm_save(img, lyr, savepath(fn), fn, 1)

    elif ext == 'tga':
        # FIXME: Last argument to file_tga_save is undocumented, not sure what
        fn = filename('RAW')
        pdb.file_tga_save(img, lyr, savepath(fn), fn, 0, 0)
        fn = filename('RLE')
        pdb.file_tga_save(img, lyr, savepath(fn), fn, 1, 0)

    elif ext == 'gif':
        fn = filename('I0')
        pdb.file_gif_save(img, lyr, savepath(fn), fn, 0, 0, 0, 0)
        fn = filename('I1')
        pdb.file_gif_save(img, lyr, savepath(fn), fn, 1, 0, 0, 0)

    elif ext == 'png':
        bits = [0, 1]
        # interlaced, bkgd block, gama block
        for i, b, g in [(i, b, g) for i in bits for b in bits for g in bits]:
            fn = filename('I{}B{}G{}'.format(i, b, g))
            pdb.file_png_save(img, lyr, savepath(fn), fn, i, 9, b, g, 1, 1, 1)

    elif ext == 'sgi':
        fn = filename('RAW')
        pdb.file_sgi_save(img, lyr, savepath(fn), fn, 0)
        fn = filename('RLE')
        pdb.file_sgi_save(img, lyr, savepath(fn), fn, 1)
        fn = filename('ARLE')
        pdb.file_sgi_save(img, lyr, savepath(fn), fn, 2)

    elif ext == 'tiff':
        fn = filename('RAW')
        pdb.file_tiff_save(img, lyr, savepath(fn), fn, 0)
        fn = filename('LZW')
        pdb.file_tiff_save(img, lyr, savepath(fn), fn, 1)
        fn = filename('PACKBITS')
        pdb.file_tiff_save(img, lyr, savepath(fn), fn, 2)
        fn = filename('DEFLATE')
        pdb.file_tiff_save(img, lyr, savepath(fn), fn, 3)

    elif ext == 'ras':
        fn = filename('RAW')
        pdb.file_sunras_save(img, lyr, savepath(fn), fn, 0)
        fn = filename('RLE')
        pdb.file_sunras_save(img, lyr, savepath(fn), fn, 1)

    else:
        fn = filename()
        pdb.gimp_file_save(img, lyr, savepath(fn), fn)


# Draw pattern on layer, helper for make_images() below
def draw_pattern(lyr, pat, alpha, direction, pixelgetter):
    assert 0 <= alpha <= 255
    assert re.match('[twxrgbycp0-9A-F]+$', pat)
    assert direction in ('x', 'y', 'width', 'height')
    dirx = direction in ('x', 'width')
    for i in range(0, len(pat)):
        pixel = pixelgetter(pat[i], alpha)
        if dirx:
            pdb.gimp_drawable_set_pixel(lyr, i, 0, len(pixel), pixel)
        else:
            pdb.gimp_drawable_set_pixel(lyr, 0, i, len(pixel), pixel)


# Create an image from the given pattern, with the specified layertype_in*,
# draw the pattern with given alpha, and save to the given extensions. Gimp
# adjust the encoder accordingly.
# * cheat for indexed formats: draw in RGB(A) and let gimp make palette
def make_images(testname, pattern, alpha, layertype_in, extensions, dirname):
    assert testname.upper() == testname
    assert len(pattern) > 0
    assert len(extensions) > 0
    assert isinstance(extensions, (list, tuple))
    assert re.match('[wxtrgbcypA-F0-9]+$', pattern)

    test_alpha = 'ALPHA' in testname or 'BINARY' in testname
    grayscale = 'GRAY' in testname

    # Indexed layer types are drawn in RGB/RGBA, and converted later
    imgtype, v0_fmtinfo = {
        GRAY_IMAGE: (GRAY, 'BPP1G'),
        GRAYA_IMAGE: (GRAY, 'BPP2GA'),
        RGB_IMAGE: (RGB, 'BPP3'),
        RGBA_IMAGE: (RGB, 'BPP4'),
        INDEXED_IMAGE: (grayscale and GRAY or RGB, 'IX'),
        INDEXEDA_IMAGE: (grayscale and GRAY or RGB, 'IXA'),
    }[layertype_in]

    # We need to supply pixels of the format of the layer
    PP = v0_pattern_pixel
    pixelgetter = {
        GRAY_IMAGE: lambda c, a: PP(c, a, 'gray'),
        GRAYA_IMAGE: lambda c, a: PP(c, a, 'graya'),
        RGB_IMAGE: lambda c, a: PP(c, a, 'rgb'),
        RGBA_IMAGE: lambda c, a: PP(c, a, 'rgba'),
        INDEXED_IMAGE: lambda c, a: PP(c, a, grayscale and 'gray' or 'rgb'),
        INDEXEDA_IMAGE: lambda c, a: PP(c, a, grayscale and 'graya' or 'rgba'),
    }[layertype_in]

    # Pick the correct layer type for indexed formats
    layertype = {
        INDEXED_IMAGE: grayscale and GRAY_IMAGE or RGB_IMAGE,
        INDEXEDA_IMAGE: grayscale and GRAYA_IMAGE or RGBA_IMAGE,
    }.get(layertype_in, layertype_in)

    # Draw pattern Nx1 and 1xN variations
    for direction in 'xy':
        # Create the gimp image, and the layer we will draw on
        w, h = (direction == 'x') and (len(pattern), 1) or (1, len(pattern))
        img = pdb.gimp_image_new(w, h, imgtype)
        lyr = pdb.gimp_layer_new(img, w, h, layertype, 'P', 100, NORMAL_MODE)

        # Add alpha layer if we are planning on encoding alpha information
        if test_alpha:
            pdb.gimp_layer_add_alpha(lyr)
            pdb.gimp_drawable_fill(lyr, TRANSPARENT_FILL)
        pdb.gimp_image_add_layer(img, lyr, 0)

        # Draw it
        draw_pattern(lyr, pattern, alpha, direction, pixelgetter)

        # Convert to indexed before saving, if needed
        if layertype_in in (INDEXED_IMAGE, INDEXEDA_IMAGE):
            colors = len(set(pattern)) + (test_alpha and 1 or 0)
            pdb.gimp_convert_indexed(img, 0, 0, colors, 0, 0, "ignored")

        # Save each individual extension
        for ext in extensions:
            save_image(dirname, img, lyr,
                       w, h, pattern, alpha, v0_fmtinfo, testname, ext)
        # FIXME: this fails?
        # pdb.gimp_image_delete(img)


# FIXME: pattern generation needs thought, this sucks..
def makepatterns(allow, include=None, exclude=None):
    src = set()
    src.update([x for x in allow])
    src.update([allow[:i] for i in range(1, len(allow) + 1)])
    for i in range(len(allow)):
        pick1, pick2 = random.choice(allow), random.choice(allow)
        src.update([pick1 + pick2])
    for i in range(3, 11) + range(14, 18) + range(31, 34):
        src.update([''.join([random.choice(allow) for k in range(i)])])
    out = []
    for srcpat in src:
        if exclude and exclude in srcpat:
            continue
        if include and include not in srcpat:
            out.append(include + srcpat[1:])
            continue
        out.append(srcpat)
    return list(set(out))


def plugin_main(dirname, do_opaque, do_binary, do_alpha):
    if not dirname:
        pdb.gimp_message("No ouput directory selected, aborting")
        return
    if not os.path.isdir(dirname) or not os.access(dirname, os.W_OK):
        pdb.gimp_message("Invalid / non-writeable output directory, aborting")
        return

    tests = []
    tests.extend({
        0: ['OPAQUE', 'GRAY-OPAQUE'],
        2: ['OPAQUE'],
        3: ['GRAY-OPAQUE'],
    }.get(do_opaque, []))
    tests.extend({
        0: ['BINARY', 'GRAY-BINARY'],
        2: ['BINARY'],
        3: ['GRAY-BINARY'],
    }.get(do_binary, []))
    tests.extend({
        0: ['ALPHA', 'GRAY-ALPHA'],
        2: ['ALPHA'],
        3: ['GRAY-ALPHA'],
    }.get(do_alpha, []))

    suite_cfg = dict(TESTSUITE_CONFIG)
    for testname, cfg in suite_cfg.items():
        if testname not in tests:
            continue
        pchars, inc, exc = cfg.pop('patterns')
        if not pchars:
            continue
        patterns = makepatterns(pchars, inc, exc)
        for alpha in cfg.pop('alpha', [255]):
            for layertype, exts in cfg.items():
                if not exts:
                    continue
                for p in patterns:
                    make_images(testname, p, alpha, layertype, exts, dirname)


register(
    proc_name="kivy_image_testsuite",
    help="Creates image test suite for Kivy ImageLoader",
    blurb=("Creates image test suite for Kivy ImageLoader. "
          "Warning: This will create thousands of images"),
    author="For kivy.org, Terje Skjaeveland",
    copyright="Copyright 2017 kivy.org (MIT license)",
    date="2017",
    imagetypes="",
    params=[
        (PF_DIRNAME, "outputdir", "Output directory:", 0),
        (PF_OPTION, "opaque", "OPAQUE tests?", 0,
                    ["All", "None", "OPAQUE", "GRAY-OPAQUE"]),
        (PF_OPTION, "binary", "BINARY tests?", 0,
                    ["All", "None", "BINARY", "GRAY-BINARY"]),
        (PF_OPTION, "alpha", "ALPHA tests?", 0,
                    ["All", "None", "ALPHA", "GRAY-ALPHA"]),
    ],
    results=[],
    function=plugin_main,
    menu="<Image>/Tools/_Kivy image testsuite...",
    label="Generate images...")

main()



import os.path
from textwrap import dedent

from kivy.app import App
from kivy.clock import Clock
from kivy import lang
from kivy.tests import GraphicUnitTest, async_run, UnitKivyApp


class AppTest(GraphicUnitTest):
    def test_start_raw_app(self):
        lang._delayed_start = None
        a = App()
        Clock.schedule_once(a.stop, .1)
        a.run()

    def test_start_app_with_kv(self):
        class TestKvApp(App):
            pass

        lang._delayed_start = None
        a = TestKvApp()
        Clock.schedule_once(a.stop, .1)
        a.run()

    def test_user_data_dir(self):
        a = App()
        data_dir = a.user_data_dir
        if not os.path.exists(data_dir):
            raise Exception("user_data_dir didnt exists")


def basic_app():
    from kivy.app import App
    from kivy.uix.label import Label

    class TestApp(UnitKivyApp, App):
        def build(self):
            return Label(text='Hello, World!')

    return TestApp()


@async_run(app_cls_func=basic_app)
async def test_basic_app(kivy_app):
    assert kivy_app.root.text == 'Hello, World!'


def button_app():
    from kivy.app import App
    from kivy.uix.togglebutton import ToggleButton

    class TestApp(UnitKivyApp, App):
        def build(self):
            return ToggleButton(text='Hello, World!')

    return TestApp()


@async_run(app_cls_func=button_app)
async def test_button_app(kivy_app):
    assert kivy_app.root.text == 'Hello, World!'
    assert kivy_app.root.state == 'normal'

    async for state, touch_pos in kivy_app.do_touch_down_up(
            widget=kivy_app.root, widget_jitter=True):
        pass

    assert kivy_app.root.state == 'down'


def scatter_app():
    from kivy.app import App
    from kivy.uix.label import Label
    from kivy.uix.scatter import Scatter

    class TestApp(UnitKivyApp, App):
        def build(self):
            label = Label(text='Hello, World!', size=('200dp', '200dp'))
            scatter = Scatter(do_scale=False, do_rotation=False)
            scatter.add_widget(label)
            return scatter

    return TestApp()


@async_run(app_cls_func=scatter_app)
async def test_drag_app(kivy_app):
    scatter = kivy_app.root
    assert tuple(scatter.pos) == (0, 0)

    async for state, touch_pos in kivy_app.do_touch_drag(
            pos=(100, 100), target_pos=(200, 200)):
        pass

    assert tuple(scatter.pos) == (100, 100)


def text_app():
    from kivy.app import App
    from kivy.uix.textinput import TextInput

    class TestApp(UnitKivyApp, App):
        def build(self):
            return TextInput()

    return TestApp()


@async_run(app_cls_func=text_app)
async def test_text_app(kivy_app):
    text = kivy_app.root
    assert text.text == ''

    # activate widget
    async for state, touch_pos in kivy_app.do_touch_down_up(widget=text):
        pass

    async for state, value in kivy_app.do_keyboard_key(key='A', num_press=4):
        pass
    async for state, value in kivy_app.do_keyboard_key(key='q', num_press=3):
        pass

    assert text.text == 'AAAAqqq'


def graphics_app():
    from kivy.app import App
    from kivy.uix.widget import Widget
    from kivy.graphics import Color, Rectangle

    class TestApp(UnitKivyApp, App):
        def build(self):
            widget = Widget()
            with widget.canvas:
                Color(1, 0, 0, 1)
                Rectangle(pos=(0, 0), size=(100, 100))
                Color(0, 1, 0, 1)
                Rectangle(pos=(100, 0), size=(100, 100))
            return widget

    return TestApp()


@async_run(app_cls_func=graphics_app)
async def test_graphics_app(kivy_app):
    widget = kivy_app.root
    (r1, g1, b1, a1), (r2, g2, b2, a2) = kivy_app.get_widget_pos_pixel(
        widget, [(50, 50), (150, 50)])

    assert not g1 and not b1 and not r2 and not b2
    assert r1 > 50 and a1 > 50 and g2 > 50 and a2 > 50


def kv_app_ref_app():
    from kivy.app import App
    from kivy.lang import Builder
    from kivy.properties import ObjectProperty
    from kivy.uix.widget import Widget

    class MyWidget(Widget):

        obj = ObjectProperty(None)

    Builder.load_string(dedent(
        """
        <MyWidget>:
            obj: app.__self__
        """))

    class TestApp(UnitKivyApp, App):
        def build(self):
            return MyWidget()

    return TestApp()


@async_run(app_cls_func=kv_app_ref_app)
async def test_leak_app_kv_property(kivy_app):
    # just tests whether the app is gc'd after the test is complete
    pass


def kv_app_default_ref_app():
    from kivy.app import App
    from kivy.lang import Builder

    class TestApp(UnitKivyApp, App):
        def build(self):
            # create property in kv and set app to it
            return Builder.load_string(dedent(
                """
                Widget:
                    obj: app.__self__
                """))

    return TestApp()


@async_run(app_cls_func=kv_app_default_ref_app)
async def test_leak_app_default_kv_property(kivy_app):
    # just tests whether the app is gc'd after the test is complete
    pass



'''
Example to show a Popup usage with the content from kv lang.
'''
from kivy.uix.popup import Popup
from kivy.uix.button import Button
from kivy.app import App
from kivy.lang import Builder

Builder.load_string('''
<CustomPopup>:
    size_hint: .5, .5
    auto_dismiss: False
    title: 'Hello world'
    Button:
        text: 'Click me to dismiss'
        on_press: root.dismiss()

''')


class CustomPopup(Popup):
    pass


class TestApp(App):
    def build(self):
        b = Button(on_press=self.show_popup, text="Show Popup")
        return b

    def show_popup(self, b):
        p = CustomPopup()
        p.open()


TestApp().run()



'''How to use Animation with RecycleView items?

In case you really want to use the Animation class with RecycleView, you'll
likely encounter an issue, as widgets are moved around, they are used to
represent different items, so an animation on a specific item is going to
affect others, and this will lead to really confusing results.

This example works around that by creating a "proxy" widget for the animation,
and, by putting it in the data, allowing the displayed widget to mimic the
animation. As the item always refers to its proxy, whichever widget is used to
display the item will keep in sync with the animation.

'''
from copy import copy

from kivy.app import App
from kivy.clock import triggered
from kivy.lang import Builder
from kivy.uix.widget import Widget
from kivy.animation import Animation
from kivy.uix.button import Button
from kivy.properties import (
    ObjectProperty, ListProperty
)


KV = '''
<Item>:
    index: None
    animation_proxy: None
    on_release: app.animate_item(self.index)


RecycleView:
    data: app.data
    viewclass: 'Item'
    RecycleBoxLayout:
        orientation: 'vertical'
        size_hint: 1, None
        height: self.minimum_height
        default_size_hint: 1, None
        default_size: 0, dp(40)
'''


class Item(Button):
    animation_proxy = ObjectProperty(allownone=True)
    _animation_proxy = None

    def update_opacity(self, proxy, opacity):
        # sync one animated property to the value in the proxy
        self.opacity = opacity

    def on_animation_proxy(self, *args):
        """When we create an animation proxy for an item, we need to bind to
        the animated property to update our own.
        """
        if self._animation_proxy:
            self._animation_proxy.unbind(opacity=self.update_opacity)

        self._animation_proxy = self.animation_proxy
        if self.animation_proxy:
            # when we are assigned an animation_proxy, sync our properties to
            # the animated version.
            self.opacity = self.animation_proxy.opacity
            self.animation_proxy.bind(opacity=self.update_opacity)
        else:
            # if we lose our animation proxy, we need to reset the animated
            # property to their default values.
            self.opacity = 1


class Application(App):
    data = ListProperty()

    def build(self):
        self.data = [
            {'index': i, 'text': 'hello {}'.format(i), 'animation_proxy': None}
            for i in range(1000)
        ]
        return Builder.load_string(KV)

    # the triggered decorator allows delaying the animation until after the
    # blue effect on the button is removed, to avoid a flash as widgets gets
    # reordered when that happens
    @triggered(timeout=0.05)
    def animate_item(self, index):
        # the animation we actually want to do on the item, note that any
        # property animated here needs to be synchronised from the proxy to the
        # animated widget (in on_animation_proxy and using methods for each
        # animation)
        proxy = Widget(opacity=1)
        item = copy(self.data[index])
        animation = (
            Animation(opacity=0, d=.1, t='out_quad')
            + Animation(opacity=1, d=5, t='out_quad')
        )
        animation.bind(on_complete=lambda *x: self.reset_animation(item))
        item['animation_proxy'] = proxy
        self.data[index] = item
        animation.start(proxy)

    def reset_animation(self, item):
        # animation is complete, widget should be garbage collected
        item['animation_proxy'] = None


if __name__ == "__main__":
    Application().run()



'''
Code Navigation Behavior
========================

The :class:`~kivy.uix.bahviors.CodeNavigationBehavior` modifies navigation
behavior in the :class:`~kivy.uix.textinput.TextInput`, making it work like an
IDE instead of a word processor.

Using this mixin gives the TextInput the ability to recognize whitespace,
punctuation and case variations (e.g. CamelCase) when moving over text. It
is currently used by the :class:`~kivy.uix.codeinput.CodeInput` widget.
'''

__all__ = ('CodeNavigationBehavior', )

from kivy.event import EventDispatcher
import string


class CodeNavigationBehavior(EventDispatcher):
    '''Code navigation behavior. Modifies the navigation behavior in TextInput
    to work like an IDE instead of a word processor. Please see the
    :mod:`code navigation behaviors module <kivy.uix.behaviors.codenavigation>`
    documentation for more information.

    .. versionadded:: 1.9.1
    '''

    def _move_cursor_word_left(self, index=None):
        pos = index or self.cursor_index()
        pos -= 1

        if pos == 0:
            return 0, 0

        col, row = self.get_cursor_from_index(pos)
        lines = self._lines

        ucase = string.ascii_uppercase
        lcase = string.ascii_lowercase
        ws = string.whitespace
        punct = string.punctuation

        mode = 'normal'

        rline = lines[row]
        c = rline[col] if len(rline) > col else '\n'
        if c in ws:
            mode = 'ws'
        elif c == '_':
            mode = 'us'
        elif c in punct:
            mode = 'punct'
        elif c not in ucase:
            mode = 'camel'

        while True:
            if col == -1:
                if row == 0:
                    return 0, 0
                row -= 1
                rline = lines[row]
                col = len(rline)
            lc = c
            c = rline[col] if len(rline) > col else '\n'
            if c == '\n':
                if lc not in ws:
                    col += 1
                break
            if mode in ('normal', 'camel') and c in ws:
                col += 1
                break
            if mode in ('normal', 'camel') and c in punct:
                col += 1
                break
            if mode == 'camel' and c in ucase:
                break
            if mode == 'punct' and (c == '_' or c not in punct):
                col += 1
                break
            if mode == 'us' and c != '_' and (c in punct or c in ws):
                col += 1
                break

            if mode == 'us' and c != '_':
                mode = ('normal' if c in ucase
                        else 'ws' if c in ws
                else 'camel')
            elif mode == 'ws' and c not in ws:
                mode = ('normal' if c in ucase
                        else 'us' if c == '_'
                else 'punct' if c in punct
                else 'camel')

            col -= 1

        if col > len(rline):
            if row == len(lines) - 1:
                return row, len(lines[row])
            row += 1
            col = 0

        return col, row

    def _move_cursor_word_right(self, index=None):
        pos = index or self.cursor_index()
        col, row = self.get_cursor_from_index(pos)
        lines = self._lines
        mrow = len(lines) - 1

        if row == mrow and col == len(lines[row]):
            return col, row

        ucase = string.ascii_uppercase
        lcase = string.ascii_lowercase
        ws = string.whitespace
        punct = string.punctuation

        mode = 'normal'

        rline = lines[row]
        c = rline[col] if len(rline) > col else '\n'
        if c in ws:
            mode = 'ws'
        elif c == '_':
            mode = 'us'
        elif c in punct:
            mode = 'punct'
        elif c in lcase:
            mode = 'camel'

        while True:
            if mode in ('normal', 'camel', 'punct') and c in ws:
                mode = 'ws'
            elif mode in ('normal', 'camel') and c == '_':
                mode = 'us'
            elif mode == 'normal' and c not in ucase:
                mode = 'camel'

            if mode == 'us':
                if c in ws:
                    mode = 'ws'
                elif c != '_':
                    break
            if mode == 'ws' and c not in ws:
                break
            if mode == 'camel' and c in ucase:
                break
            if mode == 'punct' and (c == '_' or c not in punct):
                break
            if mode != 'punct' and c != '_' and c in punct:
                break

            col += 1

            if col > len(rline):
                if row == mrow:
                    return len(rline), mrow
                row += 1
                rline = lines[row]
                col = 0

            c = rline[col] if len(rline) > col else '\n'
            if c == '\n':
                break

        return col, row



'''
uix.gridlayout tests
========================
'''

import unittest
import pytest
from kivy.tests.common import GraphicUnitTest

from kivy.uix.gridlayout import GridLayout


class GridLayoutTest(unittest.TestCase):

    def test_gridlayout_get_max_widgets_cols_rows_None(self):
        gl = GridLayout()
        expected = None
        value = gl.get_max_widgets()
        self.assertEqual(expected, value)

    def test_gridlayout_get_max_widgets_rows_None(self):
        gl = GridLayout()
        gl.cols = 1
        expected = None
        value = gl.get_max_widgets()
        self.assertEqual(expected, value)

    def test_gridlayout_get_max_widgets_cols_None(self):
        gl = GridLayout()
        gl.rows = 1
        expected = None
        value = gl.get_max_widgets()
        self.assertEqual(expected, value)

    def test_gridlayout_get_max_widgets_with_rows_cols(self):
        gl = GridLayout()
        gl.rows = 5
        gl.cols = 3
        expected = 15
        value = gl.get_max_widgets()
        self.assertEqual(expected, value)


class UixGridLayoutTest(GraphicUnitTest):

    def test_rows_cols_sizes(self):
        # ref github issue #5278 _init_rows_cols_sizes fix
        # this combination could trigger an error
        gl = GridLayout()
        gl.cols = 1
        gl.cols_minimum = {i: 10 for i in range(10)}
        gl.add_widget(GridLayout())
        self.render(gl)


@pytest.mark.parametrize(
    "n_cols, n_rows, orientation, expectation", [
        (2, 3, 'lr-tb', [(0, 0), (1, 0), (0, 1), (1, 1), (0, 2), (1, 2)]),
        (2, 3, 'lr-bt', [(0, 2), (1, 2), (0, 1), (1, 1), (0, 0), (1, 0)]),
        (2, 3, 'rl-tb', [(1, 0), (0, 0), (1, 1), (0, 1), (1, 2), (0, 2)]),
        (2, 3, 'rl-bt', [(1, 2), (0, 2), (1, 1), (0, 1), (1, 0), (0, 0)]),
        (2, 3, 'tb-lr', [(0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2)]),
        (2, 3, 'tb-rl', [(1, 0), (1, 1), (1, 2), (0, 0), (0, 1), (0, 2)]),
        (2, 3, 'bt-lr', [(0, 2), (0, 1), (0, 0), (1, 2), (1, 1), (1, 0)]),
        (2, 3, 'bt-rl', [(1, 2), (1, 1), (1, 0), (0, 2), (0, 1), (0, 0)]),
    ]
)
def test_create_idx_iter(
        n_cols, n_rows, orientation, expectation):
    from kivy.uix.gridlayout import GridLayout
    gl = GridLayout(orientation=orientation)
    index_iter = gl._create_idx_iter(n_cols, n_rows)
    assert expectation == list(index_iter)


@pytest.mark.parametrize("orientation", [
    'lr-tb', 'lr-bt', 'rl-tb', 'rl-bt',
    'tb-lr', 'tb-rl', 'bt-lr', 'bt-rl',
])
def test_create_idx_iter2(orientation):
    from kivy.uix.gridlayout import GridLayout
    gl = GridLayout(orientation=orientation)
    index_iter = gl._create_idx_iter(1, 1)
    assert [(0, 0)] == list(index_iter)


@pytest.mark.parametrize(
    "n_cols, n_rows, orientation, n_children, expectation", [
        (3, None, 'lr-tb', 4, [(0, 15), (10, 15), (20, 15), (0, 0)]),
        (3, None, 'lr-bt', 4, [(0, 0), (10, 0), (20, 0), (0, 15)]),
        (3, None, 'rl-tb', 4, [(20, 15), (10, 15), (0, 15), (20, 0)]),
        (3, None, 'rl-bt', 4, [(20, 0), (10, 0), (0, 0), (20, 15)]),
        (None, 3, 'tb-lr', 4, [(0, 20), (0, 10), (0, 0), (15, 20)]),
        (None, 3, 'tb-rl', 4, [(15, 20), (15, 10), (15, 0), (0, 20)]),
        (None, 3, 'bt-lr', 4, [(0, 0), (0, 10), (0, 20), (15, 0)]),
        (None, 3, 'bt-rl', 4, [(15, 0), (15, 10), (15, 20), (0, 0)]),
    ]
)
def test_children_pos(n_cols, n_rows, orientation, n_children, expectation):
    from kivy.uix.widget import Widget
    from kivy.uix.gridlayout import GridLayout
    gl = GridLayout(
        cols=n_cols, rows=n_rows, orientation=orientation,
        pos=(0, 0), size=(30, 30))
    for __ in range(n_children):
        gl.add_widget(Widget())
    gl.do_layout()
    actual_layout = [tuple(c.pos) for c in reversed(gl.children)]
    assert actual_layout == expectation


if __name__ == '__main__':
    unittest.main()



'''
Repeat Texture on Resize
========================

This examples repeats the letter 'K' (mtexture1.png) 64 times in a window.
You should see 8 rows and 8 columns of white K letters, along a label
showing the current size. As you resize the window, it stays an 8x8.
This example includes a label with a colored background.

Note the image mtexture1.png is a white 'K' on a transparent background, which
makes it hard to see.
'''

from kivy.app import App
from kivy.uix.image import Image
from kivy.uix.label import Label
from kivy.properties import ObjectProperty, ListProperty
from kivy.lang import Builder

kv = '''
<LabelOnBackground>:
    canvas.before:
        Color:
            rgb: self.background
        Rectangle:
            pos: self.pos
            size: self.size

FloatLayout:
    canvas.before:
        Color:
            rgb: 1, 1, 1
        Rectangle:
            pos: self.pos
            size: self.size
            texture: app.texture

    LabelOnBackground:
        text: '{} (try to resize the window)'.format(root.size)
        color: (0.4, 1, 1, 1)
        background: (.3, .3, .3)
        pos_hint: {'center_x': .5, 'center_y': .5 }
        size_hint: None, None
        height: 30
        width: 250

'''


class LabelOnBackground(Label):
    background = ListProperty((0.2, 0.2, 0.2))


class RepeatTexture(App):

    texture = ObjectProperty()

    def build(self):
        self.texture = Image(source='mtexture1.png').texture
        self.texture.wrap = 'repeat'
        self.texture.uvsize = (8, 8)
        return Builder.load_string(kv)


RepeatTexture().run()



from kivy.tests.common import GraphicUnitTest
from kivy import setupconfig


class WindowInfoTest(GraphicUnitTest):
    def test_window_info_nonzero(self):
        from kivy.core.window import Window
        window_info = Window.get_window_info()
        if window_info is None:
            return

        if setupconfig.USE_X11:
            from kivy.core.window.window_info import WindowInfoX11
            if isinstance(window_info, WindowInfoX11):
                self.assertNotEqual(window_info.display, 0)
                self.assertNotEqual(window_info.window, 0)

        if setupconfig.USE_WAYLAND:
            from kivy.core.window.window_info import WindowInfoWayland
            if isinstance(window_info, WindowInfoWayland):
                self.assertNotEqual(window_info.display, 0)
                self.assertNotEqual(window_info.surface, 0)
                self.assertNotEqual(window_info.shell_surface, 0)

        if setupconfig.PLATFORM == 'win32':
            from kivy.core.window.window_info import WindowInfoWindows
            if isinstance(window_info, WindowInfoWindows):
                self.assertNotEqual(window_info.HWND, 0)
                self.assertNotEqual(window_info.HDC, 0)



# -*- coding: utf-8 -*-

from kivy.base import runTouchApp
from kivy.lang import Builder
from kivy.factory import Factory as F


class DemoBox(F.ButtonBehavior, F.BoxLayout):
    base_direction = F.StringProperty(None, allownone=True)
    font_context = F.StringProperty(None, allownone=True)
    font_size = F.NumericProperty(10)


runTouchApp(Builder.load_string('''
#:import F kivy.factory.Factory
<MenuLabel@Label>:
    bold: True
    size_hint_x: None
    width: '150dp'
    text_size: self.size
    halign: 'right'
    valign: 'middle'
<MyToggleButton@ToggleButton>:
    hacked_state: False
    allow_no_selection: False
    on_state: self.hacked_state = self.state == 'down'
<HalignButton@MyToggleButton>:
    text: 'auto'
    group: 'halign'
    on_hacked_state: for c in root.parent.__boxes.children: \
                c.halign = root.text
<FontButton@MyToggleButton>:
    text: '20'
    group: 'font'
    on_hacked_state: for c in root.parent.__boxes.children: \
                c.font_size = int(root.text)
<ContextButton@MyToggleButton>:
    text: 'system://'
    group: 'context'
    on_hacked_state: for c in root.parent.__boxes.children: \
                c.font_context = root.text
<BaseDirButton@MyToggleButton>:
    text: 'None'
    group: 'basedir'
    on_hacked_state: for c in root.parent.__boxes.children: \
                c.base_direction = root.text
<EditPopup@Popup>:
    size_hint: .8, .8
    text: ''
    BoxLayout:
        orientation: 'vertical'
        TextInput:
            multiline: True
            id: ti
            text: root.text
            cursor_width: 3
        Button:
            size_hint_y: None
            text: 'Close'
            on_press: root.dismiss()
<DemoBox>:
    text: ti.text
    language: ''
    halign: 'auto'
    font_size: '20sp'
    font_context: 'system://'
    base_direction: None
    markup: True
    orientation: 'horizontal'
    padding: '5dp'
    spacing: '5dp'
    size_hint_y: None
    height: lbl.texture_size[1] + 25
    on_text:
        if self.text: ti.text = self.text
    Button:
        text: 'Open'
        size_hint_x: None
        on_press:
            pop = F.EditPopup()
            pop.text = lbl.text
            pop.ids.ti.halign = root.halign
            pop.ids.ti.font_size = root.font_size
            pop.ids.ti.font_context = root.font_context != 'None' \
                                      and root.font_context or None
            pop.ids.ti.base_direction = root.base_direction != 'None' \
                                        and root.base_direction or None
            pop.title = 'Edit: {}'.format(root.language)
            pop.open()
    Label:
        text: root.language
        text_size: self.size
        halign: 'left'
        bold: True
        size_hint_x: None
        width: 150
        halign: 'left'
        valign: 'middle'
        font_size: '15sp'
        color: 1, 1, 0, 1
        canvas.before:
            Color:
                rgba: 1, 1, 1, .5
            Rectangle:
                pos: self.pos
                size: self.size
    Label:
        id: lbl
        halign: root.halign
        text: root.text
        markup: root.markup
        font_context: root.font_context != 'None' and \
                      root.font_context or None
        base_direction: root.base_direction != 'None' and \
                        root.base_direction or None
        font_size: root.font_size
        text_size: self.width, None
        canvas.before:
            Color:
                rgba: 1, 1, 1, .1
            Rectangle:
                pos: self.pos
                size: self.size
    TextInput:
        id: ti
        halign: root.halign
        cursor_width: 3
        font_context: root.font_context != 'None' and root.font_context or None
        base_direction: root.base_direction != 'None' and \
                        root.base_direction or None
        font_size: root.font_size
BoxLayout:
    orientation: 'vertical'
    ScrollView:
        id: sv
        bar_color: [.1, .75, .1, .95]
        bar_inactive_color: [.5, .5, .5, .8]
        bar_width: '20dp'
        BoxLayout:
            id: boxes
            orientation: 'vertical'
            size_hint: None, None
            height: self.minimum_height
            width: sv.width - 25  # don't draw below scrollbar
            DemoBox:
                markup: True
                language: 'Arabic'
                text: 'اَلْحَمْدُ لِلّٰهِ رَبِّ \
                    [color=ff0000]الْعَالَمِينَ\
                    \\nاَلرَّحْ[/color]مٰنِ الرَّحِيمِ'
            DemoBox:
                language: 'Arabic + English'
                text: 'اِيَّاكَ نَعْبُدُ Kivyوَ اِيَّاكَ نَسْتَعِينُ\\nKivy'
            DemoBox:
                language: 'English + Arabic'
                text: 'Kivy\\nاِيَّاكَ نَعْبُدُ وَKivy اِيَّاكَ نَسْتَعِينُ'
            DemoBox:
                language: 'Hebrew'
                text: 'בְּרֵאשִׁ֖ית בָּרָ֣א \
                       \\nאֱלֹהִ֑ים אֵ֥ת הַשָּׁמַ֖יִם וְאֵ֥ת הָאָֽרֶץ׃'
            DemoBox:
                language: 'Hebrew + English'
                text: 'בְּרֵאKivyשִׁ֖ית בָּרָ֣א \\nKivy'
            DemoBox:
                language: 'English + Hebrew'
                text: 'Kivy\\nבְּרֵאשִׁ֖יKivyת בָּרָ֣א '
            DemoBox:
                language: 'Chinese'
                text: '你好，这是中文竖排测试。\\n欢迎来到中国北京。'
            DemoBox:
                language: 'Chinese + English'
                text: '你好，这是Kivy中文竖排测试。\\nKivy'
            DemoBox:
                language: 'English + Chinese'
                text: 'Kivy\\n你好，这是中Kivy文竖排测试。'
            DemoBox:
                language: 'Syriac'
                text: 'ܐܬܐܡܘܢ ܥܠܝܡܢ \\nܘܥܠܝܡܬܢ'
            DemoBox:
                language: 'Syriac + English'
                text: 'ܐܬܐܡܘܢ Kivyܥܠܝܡܢ ܘܥܠܝܡܬܢ\\nKivy'
            DemoBox:
                language: 'English + Syriac'
                text: 'Kivyܐܬܐܡܘܢ ܥܠKivyܝܡܢ ܘܥܠܝܡܬܢ\\nKivy'
            DemoBox:
                language: 'Tibetan'
                text: ' འབྲུག་རྒྱལ་ཁབ\\n་འདི་དགའ་ཏོག་ཏོ་ཡོད།'
            DemoBox:
                language: 'Tibetan + English'
                text: ' འ Kivyབྲུག་རྒྱལ་ཁབ\\n་འདི་དགKivyའ་ཏོག་ཏོ་ཡོད།'
            DemoBox:
                language: 'English + Tibetan'
                text: 'Regular letters\\n\
                        འབྲུག་རྒྱལ་ཁབ་འདི་དགའ་Kivyཏོག་ཏོ་ཡོད།'
            DemoBox:
                language: 'Braille (EN)'
                text: '⠊⠀⠉⠁⠝⠀⠑⠁⠞⠀⠛⠇⠁⠎⠎⠀\\n⠁⠝⠙⠀⠊⠞⠀⠙⠕⠑⠎⠝⠞⠀⠓⠥⠗⠞⠀⠍⠑'
            DemoBox:
                language: 'Braille (EN) + English'
                text: '⠊⠀⠉⠁⠝⠀⠑⠁⠞Kivy\\n⠀⠛⠇⠁⠎⠎⠀⠁⠝Kivy⠙⠀⠊⠞⠀⠙⠕⠑⠎⠝⠞⠀⠓⠥⠗⠞⠀⠍⠑'
            DemoBox:
                markup: True
                language: 'Braille (EN) + English'
                text: '⠊⠀⠉⠁⠝⠀⠑⠁⠞Kivy\\n⠀⠛⠇⠁⠎⠎⠀⠁⠝Kivy\
                       ⠙⠀⠊⠞⠀[color=ff0000]⠙⠕⠑⠎⠝⠞⠀⠓⠥⠗⠞⠀⠍⠑[/color]'
    # --------------------------------------- HALIGN
    BoxLayout:
        __boxes: boxes
        orientation: 'horizontal'
        size_hint_y: None
        height: '30dp'
        MenuLabel:
            text: 'set halign ='
        HalignButton:
            text: 'auto'
            state: 'down'
        HalignButton:
            text: 'left'
        HalignButton:
            text: 'center'
        HalignButton:
            text: 'right'
# Not supported by textinput
#        HalignButton:
#            text: 'justify'
    # --------------------------------BASE DIRECTION
    BoxLayout:
        __boxes: boxes
        orientation: 'horizontal'
        size_hint_y: None
        height: '30dp'
        MenuLabel:
            text: 'set base_direction ='
        BaseDirButton:
            text: 'None'
            state: 'down'
        BaseDirButton:
            text: 'ltr'
        BaseDirButton:
            text: 'rtl'
    # --------------------------------- FONT CONTEXT
    BoxLayout:
        __boxes: boxes
        orientation: 'horizontal'
        size_hint_y: None
        height: '30dp'
        MenuLabel:
            text: 'set font_context ='
        ContextButton:
            text: 'None'
        ContextButton:
            text: 'system://'
            state: 'down'
        ContextButton:
            text: 'system://user'
        ContextButton:
            text: 'user'
    # ------------------------------------ FONT SIZE
    BoxLayout:
        __boxes: boxes
        orientation: 'horizontal'
        size_hint_y: None
        height: '30dp'
        MenuLabel:
            text: 'set font_size ='
        FontButton:
            text: '10'
        FontButton:
            text: '20'
            state: 'down'
        FontButton:
            text: '30'
        FontButton:
            text: '40'
        FontButton:
            text: '50'
        FontButton:
            text: '75'
        FontButton:
            text: '100'
#    BoxLayout:
#        orientation: 'horizontal'
#        size_hint_y: None
#        height: '30dp'
#        Button:
#            color: 0, 1, 0, 1
#            text: 'Add font: user'
#        Button:
#            color: 0, 1, 0, 1
#            text: 'Add font: system://user'
#        ToggleButton:
#            text: 'Enable markup'
#            state: 'down'
#            on_state:
#                for c in boxes.children: c.markup = self.state == 'down'
'''))



from kivy.app import App
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.button import Button


class SpecialButton(Button):
    pass


class CustomLayout(BoxLayout):
    pass


class TestApp(App):
    pass


if __name__ == '__main__':
    TestApp().run()



from jnius import autoclass, PythonJavaClass, java_method
from kivy.clock import Clock
from kivy.graphics.texture import Texture
from kivy.graphics import Fbo, Callback, Rectangle
from kivy.core.camera import CameraBase
import threading


Camera = autoclass('android.hardware.Camera')
SurfaceTexture = autoclass('android.graphics.SurfaceTexture')
GL_TEXTURE_EXTERNAL_OES = autoclass(
    'android.opengl.GLES11Ext').GL_TEXTURE_EXTERNAL_OES
ImageFormat = autoclass('android.graphics.ImageFormat')


class PreviewCallback(PythonJavaClass):
    """
    Interface used to get back the preview frame of the Android Camera
    """
    __javainterfaces__ = ('android.hardware.Camera$PreviewCallback', )

    def __init__(self, callback):
        super(PreviewCallback, self).__init__()
        self._callback = callback

    @java_method('([BLandroid/hardware/Camera;)V')
    def onPreviewFrame(self, data, camera):
        self._callback(data, camera)


class CameraAndroid(CameraBase):
    """
    Implementation of CameraBase using Android API
    """

    _update_ev = None

    def __init__(self, **kwargs):
        self._android_camera = None
        self._preview_cb = PreviewCallback(self._on_preview_frame)
        self._buflock = threading.Lock()
        super(CameraAndroid, self).__init__(**kwargs)

    def __del__(self):
        self._release_camera()

    def init_camera(self):
        self._release_camera()
        self._android_camera = Camera.open(self._index)
        params = self._android_camera.getParameters()
        width, height = self._resolution
        params.setPreviewSize(width, height)
        params.setFocusMode('continuous-picture')
        self._android_camera.setParameters(params)
        # self._android_camera.setDisplayOrientation()
        self.fps = 30.

        pf = params.getPreviewFormat()
        assert(pf == ImageFormat.NV21)  # default format is NV21
        self._bufsize = int(ImageFormat.getBitsPerPixel(pf) / 8. *
                            width * height)

        self._camera_texture = Texture(width=width, height=height,
                                       target=GL_TEXTURE_EXTERNAL_OES,
                                       colorfmt='rgba')
        self._surface_texture = SurfaceTexture(int(self._camera_texture.id))
        self._android_camera.setPreviewTexture(self._surface_texture)

        self._fbo = Fbo(size=self._resolution)
        self._fbo['resolution'] = (float(width), float(height))
        self._fbo.shader.fs = '''
            #extension GL_OES_EGL_image_external : require
            #ifdef GL_ES
                precision highp float;
            #endif

            /* Outputs from the vertex shader */
            varying vec4 frag_color;
            varying vec2 tex_coord0;

            /* uniform texture samplers */
            uniform sampler2D texture0;
            uniform samplerExternalOES texture1;
            uniform vec2 resolution;

            void main()
            {
                vec2 coord = vec2(tex_coord0.y * (
                    resolution.y / resolution.x), 1. -tex_coord0.x);
                gl_FragColor = texture2D(texture1, tex_coord0);
            }
        '''
        with self._fbo:
            self._texture_cb = Callback(lambda instr:
                                        self._camera_texture.bind)
            Rectangle(size=self._resolution)

    def _release_camera(self):
        if self._android_camera is None:
            return

        self.stop()
        self._android_camera.release()
        self._android_camera = None

        # clear texture and it'll be reset in `_update` pointing to new FBO
        self._texture = None
        del self._fbo, self._surface_texture, self._camera_texture

    def _on_preview_frame(self, data, camera):
        with self._buflock:
            if self._buffer is not None:
                # add buffer back for reuse
                self._android_camera.addCallbackBuffer(self._buffer)
            self._buffer = data
        # check if frame grabbing works
        # print self._buffer, len(self.frame_data)

    def _refresh_fbo(self):
        self._texture_cb.ask_update()
        self._fbo.draw()

    def start(self):
        super(CameraAndroid, self).start()

        with self._buflock:
            self._buffer = None
        for k in range(2):  # double buffer
            buf = b'\x00' * self._bufsize
            self._android_camera.addCallbackBuffer(buf)
        self._android_camera.setPreviewCallbackWithBuffer(self._preview_cb)

        self._android_camera.startPreview()
        if self._update_ev is not None:
            self._update_ev.cancel()
        self._update_ev = Clock.schedule_interval(self._update, 1 / self.fps)

    def stop(self):
        super(CameraAndroid, self).stop()
        if self._update_ev is not None:
            self._update_ev.cancel()
            self._update_ev = None
        self._android_camera.stopPreview()

        self._android_camera.setPreviewCallbackWithBuffer(None)
        # buffer queue cleared as well, to be recreated on next start
        with self._buflock:
            self._buffer = None

    def _update(self, dt):
        self._surface_texture.updateTexImage()
        self._refresh_fbo()
        if self._texture is None:
            self._texture = self._fbo.texture
            self.dispatch('on_load')
        self._copy_to_gpu()

    def _copy_to_gpu(self):
        """
        A dummy placeholder (the image is already in GPU) to be consistent
        with other providers.
        """
        self.dispatch('on_texture')

    def grab_frame(self):
        """
        Grab current frame (thread-safe, minimal overhead)
        """
        with self._buflock:
            if self._buffer is None:
                return None
            buf = self._buffer.tostring()
            return buf

    def decode_frame(self, buf):
        """
        Decode image data from grabbed frame.

        This method depends on OpenCV and NumPy - however it is only used for
        fetching the current frame as a NumPy array, and not required when
        this :class:`CameraAndroid` provider is simply used by a
        :class:`~kivy.uix.camera.Camera` widget.
        """
        import numpy as np
        from cv2 import cvtColor

        w, h = self._resolution
        arr = np.fromstring(buf, 'uint8').reshape((h + h / 2, w))
        arr = cvtColor(arr, 93)  # NV21 -> BGR
        return arr

    def read_frame(self):
        """
        Grab and decode frame in one call
        """
        return self.decode_frame(self.grab_frame())

    @staticmethod
    def get_camera_count():
        """
        Get the number of available cameras.
        """
        return Camera.getNumberOfCameras()



from kivy.tests import async_run, UnitKivyApp


def dropdown_app():
    from kivy.app import App
    from kivy.uix.button import Button
    from kivy.uix.dropdown import DropDown
    from kivy.uix.label import Label

    class RootButton(Button):

        dropdown = None

        def on_touch_down(self, touch):
            assert self.dropdown.attach_to is None
            return super(RootButton, self).on_touch_down(touch)

        def on_touch_move(self, touch):
            assert self.dropdown.attach_to is None
            return super(RootButton, self).on_touch_move(touch)

        def on_touch_up(self, touch):
            assert self.dropdown.attach_to is None
            return super(RootButton, self).on_touch_up(touch)

    class TestApp(UnitKivyApp, App):
        def build(self):
            root = RootButton(text='Root')
            self.attach_widget = Label(text='Attached widget')
            root.add_widget(self.attach_widget)

            root.dropdown = self.dropdown = DropDown(
                auto_dismiss=True, min_state_time=0)
            self.inner_widget = w = Label(
                size_hint=(None, None), text='Dropdown')
            root.dropdown.add_widget(w)
            return root

    return TestApp()


@async_run(app_cls_func=dropdown_app)
async def test_dropdown_app(kivy_app):
    await kivy_app.wait_clock_frames(2)
    dropdown = kivy_app.dropdown
    button = kivy_app.root
    assert dropdown.attach_to is None

    kivy_app.attach_widget.size = button.width * 3 / 5, 2
    kivy_app.attach_widget.top = button.height
    kivy_app.inner_widget.size = button.width * 3 / 5, button.height

    # just press button
    async for _ in kivy_app.do_touch_down_up(widget=button):
        pass
    async for _ in kivy_app.do_touch_drag(widget=button, dx=button.width / 4):
        pass

    # open dropdown
    dropdown.open(kivy_app.attach_widget)
    dropdown.pos = 0, 0
    await kivy_app.wait_clock_frames(2)
    assert dropdown.attach_to is not None

    # press within dropdown area - should stay open
    async for _ in kivy_app.do_touch_down_up(widget=button):
        pass
    assert dropdown.attach_to is not None
    # start in dropdown but release outside - should stay open
    async for _ in kivy_app.do_touch_drag(widget=button, dx=button.width / 4):
        pass
    assert dropdown.attach_to is not None

    # start outside but release in dropdown - should close
    async for _ in kivy_app.do_touch_drag(
            pos=(button.center_x + button.width / 4, button.center_y),
            target_widget=button):
        pass
    assert dropdown.attach_to is None

    # open dropdown again
    dropdown.open(kivy_app.attach_widget)
    dropdown.pos = 0, 0
    await kivy_app.wait_clock_frames(2)
    assert dropdown.attach_to is not None

    # press outside dropdown area to close it - should close
    async for _ in kivy_app.do_touch_down_up(
            pos=(button.center_x + button.width / 4, button.center_y)):
        pass
    assert dropdown.attach_to is None



'''
FBO Canvas
==========

This demonstrates a layout using an FBO (Frame Buffer Off-screen)
instead of a plain canvas. You should see a black canvas with a
button labelled 'FBO' in the bottom left corner. Clicking it
animates the button moving right to left.
'''

__all__ = ('FboFloatLayout', )

from kivy.graphics import Color, Rectangle, Canvas, ClearBuffers, ClearColor
from kivy.graphics.fbo import Fbo
from kivy.uix.floatlayout import FloatLayout
from kivy.properties import ObjectProperty, NumericProperty
from kivy.app import App
from kivy.core.window import Window
from kivy.animation import Animation
from kivy.factory import Factory


class FboFloatLayout(FloatLayout):

    texture = ObjectProperty(None, allownone=True)

    alpha = NumericProperty(1)

    def __init__(self, **kwargs):
        self.canvas = Canvas()
        with self.canvas:
            self.fbo = Fbo(size=self.size)
            self.fbo_color = Color(1, 1, 1, 1)
            self.fbo_rect = Rectangle()

        with self.fbo:
            ClearColor(0, 0, 0, 0)
            ClearBuffers()

        # wait that all the instructions are in the canvas to set texture
        self.texture = self.fbo.texture
        super(FboFloatLayout, self).__init__(**kwargs)

    def add_widget(self, *largs):
        # trick to attach graphics instruction to fbo instead of canvas
        canvas = self.canvas
        self.canvas = self.fbo
        ret = super(FboFloatLayout, self).add_widget(*largs)
        self.canvas = canvas
        return ret

    def remove_widget(self, *largs):
        canvas = self.canvas
        self.canvas = self.fbo
        super(FboFloatLayout, self).remove_widget(*largs)
        self.canvas = canvas

    def on_size(self, instance, value):
        self.fbo.size = value
        self.texture = self.fbo.texture
        self.fbo_rect.size = value

    def on_pos(self, instance, value):
        self.fbo_rect.pos = value

    def on_texture(self, instance, value):
        self.fbo_rect.texture = value

    def on_alpha(self, instance, value):
        self.fbo_color.rgba = (1, 1, 1, value)


class ScreenLayerApp(App):
    def build(self):

        f = FboFloatLayout()
        b = Factory.Button(text="FBO", size_hint=(None, None))
        f.add_widget(b)

        def anim_btn(*args):
            if b.pos[0] == 0:
                Animation(x=f.width - b.width).start(b)
            else:
                Animation(x=0).start(b)
        b.bind(on_press=anim_btn)

        return f


if __name__ == "__main__":
    ScreenLayerApp().run()



'''
Touch Ripple
============

.. versionadded:: 1.10.1

.. warning::
    This code is still experimental, and its API is subject to change in a
    future version.

This module contains `mixin <https://en.wikipedia.org/wiki/Mixin>`_ classes
to add a touch ripple visual effect known from `Google Material Design
<https://en.wikipedia.org/wiki/Material_Design>_` to widgets.

For an overview of behaviors, please refer to the :mod:`~kivy.uix.behaviors`
documentation.

The class :class:`~kivy.uix.behaviors.touchripple.TouchRippleBehavior` provides
rendering the ripple animation.

The class :class:`~kivy.uix.behaviors.touchripple.TouchRippleButtonBehavior`
basically provides the same functionality as
:class:`~kivy.uix.behaviors.button.ButtonBehavior` but rendering the ripple
animation instead of default press/release visualization.
'''
from kivy.animation import Animation
from kivy.clock import Clock
from kivy.graphics import CanvasBase, Color, Ellipse, ScissorPush, ScissorPop
from kivy.properties import BooleanProperty, ListProperty, NumericProperty, \
    ObjectProperty, StringProperty
from kivy.uix.relativelayout import RelativeLayout


__all__ = (
    'TouchRippleBehavior',
    'TouchRippleButtonBehavior'
)


class TouchRippleBehavior(object):
    '''Touch ripple behavior.

    Supposed to be used as mixin on widget classes.

    Ripple behavior does not trigger automatically, concrete implementation
    needs to call :func:`ripple_show` respective :func:`ripple_fade` manually.

    Example
    -------

    Here we create a Label which renders the touch ripple animation on
    interaction::

        class RippleLabel(TouchRippleBehavior, Label):

            def __init__(self, **kwargs):
                super(RippleLabel, self).__init__(**kwargs)

            def on_touch_down(self, touch):
                collide_point = self.collide_point(touch.x, touch.y)
                if collide_point:
                    touch.grab(self)
                    self.ripple_show(touch)
                    return True
                return False

            def on_touch_up(self, touch):
                if touch.grab_current is self:
                    touch.ungrab(self)
                    self.ripple_fade()
                    return True
                return False
    '''

    ripple_rad_default = NumericProperty(10)
    '''Default radius the animation starts from.

    :attr:`ripple_rad_default` is a :class:`~kivy.properties.NumericProperty`
    and defaults to `10`.
    '''

    ripple_duration_in = NumericProperty(.5)
    '''Animation duration taken to show the overlay.

    :attr:`ripple_duration_in` is a :class:`~kivy.properties.NumericProperty`
    and defaults to `0.5`.
    '''

    ripple_duration_out = NumericProperty(.2)
    '''Animation duration taken to fade the overlay.

    :attr:`ripple_duration_out` is a :class:`~kivy.properties.NumericProperty`
    and defaults to `0.2`.
    '''

    ripple_fade_from_alpha = NumericProperty(.5)
    '''Alpha channel for ripple color the animation starts with.

    :attr:`ripple_fade_from_alpha` is a
    :class:`~kivy.properties.NumericProperty` and defaults to `0.5`.
    '''

    ripple_fade_to_alpha = NumericProperty(.8)
    '''Alpha channel for ripple color the animation targets to.

    :attr:`ripple_fade_to_alpha` is a :class:`~kivy.properties.NumericProperty`
    and defaults to `0.8`.
    '''

    ripple_scale = NumericProperty(2.)
    '''Max scale of the animation overlay calculated from max(width/height) of
    the decorated widget.

    :attr:`ripple_scale` is a :class:`~kivy.properties.NumericProperty`
    and defaults to `2.0`.
    '''

    ripple_func_in = StringProperty('in_cubic')
    '''Animation callback for showing the overlay.

    :attr:`ripple_func_in` is a :class:`~kivy.properties.StringProperty`
    and defaults to `in_cubic`.
    '''

    ripple_func_out = StringProperty('out_quad')
    '''Animation callback for hiding the overlay.

    :attr:`ripple_func_out` is a :class:`~kivy.properties.StringProperty`
    and defaults to `out_quad`.
    '''

    ripple_rad = NumericProperty(10)
    ripple_pos = ListProperty([0, 0])
    ripple_color = ListProperty((1., 1., 1., .5))

    def __init__(self, **kwargs):
        super(TouchRippleBehavior, self).__init__(**kwargs)
        self.ripple_pane = CanvasBase()
        self.canvas.add(self.ripple_pane)
        self.bind(
            ripple_color=self._ripple_set_color,
            ripple_pos=self._ripple_set_ellipse,
            ripple_rad=self._ripple_set_ellipse
        )
        self.ripple_ellipse = None
        self.ripple_col_instruction = None

    def ripple_show(self, touch):
        '''Begin ripple animation on current widget.

        Expects touch event as argument.
        '''
        Animation.cancel_all(self, 'ripple_rad', 'ripple_color')
        self._ripple_reset_pane()
        x, y = self.to_window(*self.pos)
        width, height = self.size
        if isinstance(self, RelativeLayout):
            self.ripple_pos = ripple_pos = (touch.x - x, touch.y - y)
        else:
            self.ripple_pos = ripple_pos = (touch.x, touch.y)
        rc = self.ripple_color
        ripple_rad = self.ripple_rad
        self.ripple_color = [rc[0], rc[1], rc[2], self.ripple_fade_from_alpha]
        with self.ripple_pane:
            ScissorPush(
                x=int(round(x)),
                y=int(round(y)),
                width=int(round(width)),
                height=int(round(height))
            )
            self.ripple_col_instruction = Color(rgba=self.ripple_color)
            self.ripple_ellipse = Ellipse(
                size=(ripple_rad, ripple_rad),
                pos=(
                    ripple_pos[0] - ripple_rad / 2.,
                    ripple_pos[1] - ripple_rad / 2.
                )
            )
            ScissorPop()
        anim = Animation(
            ripple_rad=max(width, height) * self.ripple_scale,
            t=self.ripple_func_in,
            ripple_color=[rc[0], rc[1], rc[2], self.ripple_fade_to_alpha],
            duration=self.ripple_duration_in
        )
        anim.start(self)

    def ripple_fade(self):
        '''Finish ripple animation on current widget.
        '''
        Animation.cancel_all(self, 'ripple_rad', 'ripple_color')
        width, height = self.size
        rc = self.ripple_color
        duration = self.ripple_duration_out
        anim = Animation(
            ripple_rad=max(width, height) * self.ripple_scale,
            ripple_color=[rc[0], rc[1], rc[2], 0.],
            t=self.ripple_func_out,
            duration=duration
        )
        anim.bind(on_complete=self._ripple_anim_complete)
        anim.start(self)

    def _ripple_set_ellipse(self, instance, value):
        ellipse = self.ripple_ellipse
        if not ellipse:
            return
        ripple_pos = self.ripple_pos
        ripple_rad = self.ripple_rad
        ellipse.size = (ripple_rad, ripple_rad)
        ellipse.pos = (
            ripple_pos[0] - ripple_rad / 2.,
            ripple_pos[1] - ripple_rad / 2.
        )

    def _ripple_set_color(self, instance, value):
        if not self.ripple_col_instruction:
            return
        self.ripple_col_instruction.rgba = value

    def _ripple_anim_complete(self, anim, instance):
        self._ripple_reset_pane()

    def _ripple_reset_pane(self):
        self.ripple_rad = self.ripple_rad_default
        self.ripple_pane.clear()


class TouchRippleButtonBehavior(TouchRippleBehavior):
    '''
    This `mixin <https://en.wikipedia.org/wiki/Mixin>`_ class provides
    a similar behavior to :class:`~kivy.uix.behaviors.button.ButtonBehavior`
    but provides touch ripple animation instead of button pressed/released as
    visual effect.

    :Events:
        `on_press`
            Fired when the button is pressed.
        `on_release`
            Fired when the button is released (i.e. the touch/click that
            pressed the button goes away).
    '''

    last_touch = ObjectProperty(None)
    '''Contains the last relevant touch received by the Button. This can
    be used in `on_press` or `on_release` in order to know which touch
    dispatched the event.

    :attr:`last_touch` is a :class:`~kivy.properties.ObjectProperty` and
    defaults to `None`.
    '''

    always_release = BooleanProperty(False)
    '''This determines whether or not the widget fires an `on_release` event if
    the touch_up is outside the widget.

    :attr:`always_release` is a :class:`~kivy.properties.BooleanProperty` and
    defaults to `False`.
    '''

    def __init__(self, **kwargs):
        self.register_event_type('on_press')
        self.register_event_type('on_release')
        super(TouchRippleButtonBehavior, self).__init__(**kwargs)

    def on_touch_down(self, touch):
        if super(TouchRippleButtonBehavior, self).on_touch_down(touch):
            return True
        if touch.is_mouse_scrolling:
            return False
        if not self.collide_point(touch.x, touch.y):
            return False
        if self in touch.ud:
            return False
        touch.grab(self)
        touch.ud[self] = True
        self.last_touch = touch
        self.ripple_show(touch)
        self.dispatch('on_press')
        return True

    def on_touch_move(self, touch):
        if touch.grab_current is self:
            return True
        if super(TouchRippleButtonBehavior, self).on_touch_move(touch):
            return True
        return self in touch.ud

    def on_touch_up(self, touch):
        if touch.grab_current is not self:
            return super(TouchRippleButtonBehavior, self).on_touch_up(touch)
        assert(self in touch.ud)
        touch.ungrab(self)
        self.last_touch = touch
        if self.disabled:
            return
        self.ripple_fade()
        if not self.always_release and not self.collide_point(*touch.pos):
            return

        # defer on_release until ripple_fade has completed
        def defer_release(dt):
            self.dispatch('on_release')
        Clock.schedule_once(defer_release, self.ripple_duration_out)
        return True

    def on_disabled(self, instance, value):
        # ensure ripple animation completes if disabled gets set to True
        if value:
            self.ripple_fade()
        return super(TouchRippleButtonBehavior, self).on_disabled(
            instance, value)

    def on_press(self):
        pass

    def on_release(self):
        pass



from kivy.lang import Builder
from kivy.uix.label import Label
from kivy.base import runTouchApp
from kivy.uix.boxlayout import BoxLayout

import os

Builder.load_string('''
#:import Clipboard kivy.core.clipboard.Clipboard
<Clip>:
    orientation: 'vertical'
    GridLayout:
        cols: 3
        size_hint_y: None
        height: self.minimum_height
        Button:
            text: 'Paste raw'
            size_hint_y: None
            height: 60
            on_release: root.make_labels(Clipboard.paste())

        Button:
            text: 'Paste & format'
            size_hint_y: None
            height: 60
            on_release: root.make_pretty_labels(Clipboard.paste())

        Button:
            text: 'Remove widgets'
            size_hint_y: None
            height: 60
            on_release: container.clear_widgets()

    ScrollView:
        GridLayout:
            cols: 1
            id: container
            size_hint_y: None
            height: self.minimum_height
''')


class Clip(BoxLayout):
    def make_labels(self, values):
        """Creates widgets from raw clipboard i.e. for each character in the
        list that is provided by Clipboard.paste()
        """
        print(repr(values))
        for value in values:
            label = Label(text=value, size_hint_y=None, height=30)
            self.ids.container.add_widget(label)

    def make_pretty_labels(self, values):
        """Creates widgets from a list of values made by splitting clipboard
        by the default OS line separator. Useful when copying columns of data.
        """
        print(repr(values))
        for value in values.split(os.linesep):
            label = Label(text=value, size_hint_y=None, height=30)
            self.ids.container.add_widget(label)


runTouchApp(Clip())



'''
GstPlayer
=========

.. versionadded:: 1.8.0

`GstPlayer` is a media player implemented specifically for Kivy with Gstreamer
1.0. It doesn't use Gi at all and is focused on what we want: the ability
to read video and stream the image in a callback, or read an audio file.
Don't use it directly but use our Core providers instead.

This player is automatically compiled if you have `pkg-config --libs --cflags
gstreamer-1.0` working.

.. warning::

    This is an external library and Kivy does not provide any support for it.
    It might change in the future and we advise you don't rely on it in your
    code.
'''

import os
if 'KIVY_DOC' in os.environ:
    GstPlayer = get_gst_version = glib_iteration = None
else:
    from kivy.lib.gstplayer._gstplayer import (
        GstPlayer, get_gst_version, glib_iteration)



'''
Audio tests
===========
'''

import unittest
import os
import pytest
if os.environ.get('KIVY_TEST_AUDIO') == '0':
    pytestmark = pytest.mark.skip("Audio is not available")

SAMPLE_FILE = os.path.join(os.path.dirname(__file__), 'sample1.ogg')
SAMPLE_LENGTH = 1.402
DELTA = SAMPLE_LENGTH * 0.01
DELAY = 0.2


class AudioTestCase(unittest.TestCase):

    def get_sound(self):
        import os
        assert os.path.exists(SAMPLE_FILE)
        from kivy.core import audio
        return audio.SoundLoader.load(SAMPLE_FILE)

    def test_length_simple(self):
        sound = self.get_sound()
        volume = sound.volume = 0.75
        length = sound.length
        self.assertAlmostEqual(SAMPLE_LENGTH, length, delta=DELTA)
        # ensure that the gstreamer play/stop doesn't mess up the volume
        assert volume == sound.volume

    def test_length_playing(self):
        import time
        sound = self.get_sound()
        sound.play()
        try:
            time.sleep(DELAY)
            length = sound.length
            self.assertAlmostEqual(SAMPLE_LENGTH, length, delta=DELTA)
        finally:
            sound.stop()
        self.assertAlmostEqual(SAMPLE_LENGTH, length, delta=DELTA)

    def test_length_stopped(self):
        import time
        sound = self.get_sound()
        sound.play()
        try:
            time.sleep(DELAY)
        finally:
            sound.stop()
        length = sound.length
        self.assertAlmostEqual(SAMPLE_LENGTH, length, delta=DELTA)


class AudioGstreamerTestCase(AudioTestCase):

    def make_sound(self, source):
        from kivy.core.audio import audio_gstreamer
        return audio_gstreamer.SoundGstreamer(source)


class AudioPygameTestCase(AudioTestCase):

    def make_sound(self, source):
        from kivy.core.audio import audio_pygame
        return audio_pygame.SoundPygame(source)



# c definition
cdef int c_GLCAP_BGRA = 0x0001
cdef int c_GLCAP_NPOT = 0x0002
cdef int c_GLCAP_S3TC = 0x0003
cdef int c_GLCAP_DXT1 = 0x0004
cdef int c_GLCAP_PVRTC = 0x0005
cdef int c_GLCAP_ETC1 = 0x0006
cdef int c_GLCAP_UNPACK_SUBIMAGE = 0x0007

# for python export
GLCAP_BGRA = c_GLCAP_NPOT
GLCAP_NPOT = c_GLCAP_NPOT
GLCAP_S3TC = c_GLCAP_S3TC
GLCAP_DXT1 = c_GLCAP_DXT1
GLCAP_PVRTC = c_GLCAP_PVRTC
GLCAP_ETC1 = c_GLCAP_ETC1
GLCAP_UNPACK_SUBIMAGE = c_GLCAP_UNPACK_SUBIMAGE



import pytest


@pytest.mark.parametrize('prop_name', (
    '_fills_row_first',
    '_fills_from_left_to_right',
    '_fills_from_top_to_bottom',
))
def test_a_certain_properties_from_the_super_class_are_overwritten(prop_name):
    from kivy.uix.bubble import Bubble
    from kivy.uix.gridlayout import GridLayout
    assert issubclass(Bubble, GridLayout)
    assert getattr(Bubble, prop_name) is not getattr(GridLayout, prop_name)


@pytest.mark.parametrize('orientation', ('vertical', 'horizontal'))
def test_always_lr_tb(orientation):
    from kivy.uix.bubble import Bubble
    b = Bubble(orientation=orientation)
    assert b._fills_row_first
    assert b._fills_from_left_to_right
    assert b._fills_from_top_to_bottom



'''
Label mipmap
============

This show how to create a mipmapped label, and the visual difference between a
non mipmapped and mipmapped label.
'''

import kivy
kivy.require('1.0.7')

from kivy.app import App
from kivy.uix.scatter import ScatterPlane
from kivy.uix.label import Label


class LabelMipmapTest(App):
    def build(self):
        s = ScatterPlane(scale=.5)
        l1 = Label(text='Kivy rulz', font_size=98, pos=(400, 100), mipmap=True)
        l2 = Label(text='Kivy rulz', font_size=98, pos=(400, 328))
        s.add_widget(l1)
        s.add_widget(l2)
        return s


if __name__ == '__main__':
    LabelMipmapTest().run()



from kivy.app import App
from kivy.uix.widget import Widget


class MyPaintWidget(Widget):
    pass


class MyPaintApp(App):
    def build(self):
        return MyPaintWidget()


if __name__ == '__main__':
    MyPaintApp().run()



import gc


def test_weak_method_on_obj():
    from kivy.weakmethod import WeakMethod

    class SomeClass:

        def do_something(self):
            pass

    obj = SomeClass()
    weak_method = WeakMethod(obj.do_something)

    assert not weak_method.is_dead()
    assert weak_method() == obj.do_something
    assert weak_method == WeakMethod(obj.do_something)
    assert weak_method != WeakMethod(SomeClass().do_something)

    del obj
    gc.collect()

    assert weak_method.is_dead()
    assert weak_method() is None
    assert weak_method != WeakMethod(SomeClass().do_something)


def test_weak_method_func():
    from kivy.weakmethod import WeakMethod

    def do_something():
        pass

    weak_method = WeakMethod(do_something)

    assert not weak_method.is_dead()
    assert weak_method() == do_something
    assert weak_method == WeakMethod(do_something)

    del do_something
    gc.collect()

    assert not weak_method.is_dead()
    assert weak_method() is not None



cdef int GI_NOOP         = 1 << 0
cdef int GI_IGNORE       = 1 << 1
cdef int GI_NEEDS_UPDATE = 1 << 2
cdef int GI_GROUP        = 1 << 3
cdef int GI_CONTEXT_MOD  = 1 << 4
cdef int GI_VERTEX_DATA  = 1 << 5
cdef int GI_COMPILER	 = 1 << 6
cdef int GI_NO_APPLY_ONCE = 1 << 7
cdef int GI_NO_REMOVE    = 1 << 8




'''
EGL Rpi Window: EGL Window provider, specialized for the Pi

Inspired by: rpi_vid_core + JF002 rpi kivy  repo
'''

__all__ = ('WindowEglRpi', )

from kivy.logger import Logger
from kivy.core.window import WindowBase
from kivy.base import EventLoop, ExceptionManager, stopTouchApp
from kivy.lib.vidcore_lite import bcm, egl
from os import environ

# Default display IDs.
(DISPMANX_ID_MAIN_LCD,
 DISPMANX_ID_AUX_LCD,
 DISPMANX_ID_HDMI,
 DISPMANX_ID_SDTV,
 DISPMANX_ID_FORCE_LCD,
 DISPMANX_ID_FORCE_TV,
 DISPMANX_ID_FORCE_OTHER) = range(7)


class WindowEglRpi(WindowBase):

    _rpi_dispmanx_id = int(environ.get("KIVY_BCM_DISPMANX_ID", "0"))
    _rpi_dispmanx_layer = int(environ.get("KIVY_BCM_DISPMANX_LAYER", "0"))

    gl_backends_ignored = ['sdl2']

    def create_window(self):
        bcm.host_init()

        w, h = bcm.graphics_get_display_size(self._rpi_dispmanx_id)
        Logger.debug('Window: Actual display size: {}x{}'.format(
            w, h))
        self._size = w, h
        self._create_window(w, h)
        self._create_egl_context(self.win, 0)
        super(WindowEglRpi, self).create_window()

    def _create_window(self, w, h):
        dst = bcm.Rect(0, 0, w, h)
        src = bcm.Rect(0, 0, w << 16, h << 16)
        display = egl.bcm_display_open(self._rpi_dispmanx_id)
        update = egl.bcm_update_start(0)
        element = egl.bcm_element_add(
            update, display, self._rpi_dispmanx_layer, dst, src)
        self.win = egl.NativeWindow(element, w, h)
        egl.bcm_update_submit_sync(update)

    def _create_egl_context(self, win, flags):
        api = egl._constants.EGL_OPENGL_ES_API
        c = egl._constants

        attribs = [
            c.EGL_RED_SIZE, 8,
            c.EGL_GREEN_SIZE, 8,
            c.EGL_BLUE_SIZE, 8,
            c.EGL_ALPHA_SIZE, 8,
            c.EGL_DEPTH_SIZE, 16,
            c.EGL_STENCIL_SIZE, 8,
            c.EGL_SURFACE_TYPE, c.EGL_WINDOW_BIT,
            c.EGL_NONE]

        attribs_context = [c.EGL_CONTEXT_CLIENT_VERSION, 2, c.EGL_NONE]

        display = egl.GetDisplay(c.EGL_DEFAULT_DISPLAY)
        egl.Initialise(display)
        egl.BindAPI(c.EGL_OPENGL_ES_API)
        egl.GetConfigs(display)
        config = egl.ChooseConfig(display, attribs, 1)[0]
        surface = egl.CreateWindowSurface(display, config, win)
        context = egl.CreateContext(display, config, None, attribs_context)
        egl.MakeCurrent(display, surface, surface, context)

        self.egl_info = (display, surface, context)
        egl.MakeCurrent(display, surface, surface, context)

    def close(self):
        egl.Terminate(self.egl_info[0])

    def flip(self):
        if not EventLoop.quit:
            egl.SwapBuffers(self.egl_info[0], self.egl_info[1])



'''
RecycleView Layouts
===================

.. versionadded:: 1.10.0

The Layouts handle the presentation of views for the
:class:`~kivy.uix.recycleview.RecycleView`.

.. warning::
    This module is highly experimental, its API may change in the future and
    the documentation is not complete at this time.
'''
from kivy.compat import string_types
from kivy.factory import Factory
from kivy.properties import StringProperty, ObjectProperty
from kivy.uix.behaviors import CompoundSelectionBehavior
from kivy.uix.recycleview.views import RecycleDataViewBehavior, \
    _view_base_cache


class LayoutChangeException(Exception):
    pass


class LayoutSelectionBehavior(CompoundSelectionBehavior):
    '''The :class:`LayoutSelectionBehavior` can be combined with
    :class:`RecycleLayoutManagerBehavior` to allow its derived classes
    selection behaviors similarly to how
    :class:`~kivy.uix.behaviors.compoundselection.CompoundSelectionBehavior`
    can be used to add selection behaviors to normal layout.

    :class:`RecycleLayoutManagerBehavior` manages its children
    differently than normal layouts or widgets so this class adapts
    :class:`~kivy.uix.behaviors.compoundselection.CompoundSelectionBehavior`
    based selection to work with :class:`RecycleLayoutManagerBehavior` as well.

    Similarly to
    :class:`~kivy.uix.behaviors.compoundselection.CompoundSelectionBehavior`,
    one can select using the keyboard or touch, which calls :meth:`select_node`
    or :meth:`deselect_node`, or one can call these methods directly. When a
    item is selected or deselected :meth:`apply_selection` is called. See
    :meth:`apply_selection`.


    '''

    key_selection = StringProperty(None, allownone=True)
    '''The key used to check whether a view of a data item can be selected
    with touch or the keyboard.

    :attr:`key_selection` is the key in data, which if present and ``True``
    will enable selection for this item from the keyboard or with a touch.
    When None, the default, not item will be selectable.

    :attr:`key_selection` is a :class:`StringProperty` and defaults to None.

    .. note::
        All data items can be selected directly using :meth:`select_node` or
        :meth:`deselect_node`, even if :attr:`key_selection` is False.
    '''

    _selectable_nodes = []
    _nodes_map = {}

    def __init__(self, **kwargs):
        self.nodes_order_reversed = False
        super(LayoutSelectionBehavior, self).__init__(**kwargs)

    def compute_sizes_from_data(self, data, flags):
        # overwrite this method so that when data changes we update
        # selectable nodes.
        key = self.key_selection
        if key is None:
            nodes = self._selectable_nodes = []
        else:
            nodes = self._selectable_nodes = [
                i for i, d in enumerate(data) if d.get(key)]

        self._nodes_map = {v: k for k, v in enumerate(nodes)}
        return super(LayoutSelectionBehavior, self).compute_sizes_from_data(
            data, flags)

    def get_selectable_nodes(self):
        # the indices of the data is used as the nodes
        return self._selectable_nodes

    def get_index_of_node(self, node, selectable_nodes):
        # the indices of the data is used as the nodes, so node
        return self._nodes_map[node]

    def goto_node(self, key, last_node, last_node_idx):
        node, idx = super(LayoutSelectionBehavior, self).goto_node(
            key, last_node, last_node_idx)
        if node is not last_node:
            self.goto_view(node)
        return node, idx

    def select_node(self, node):
        if super(LayoutSelectionBehavior, self).select_node(node):
            view = self.recycleview.view_adapter.get_visible_view(node)
            if view is not None:
                self.apply_selection(node, view, True)

    def deselect_node(self, node):
        if super(LayoutSelectionBehavior, self).deselect_node(node):
            view = self.recycleview.view_adapter.get_visible_view(node)
            if view is not None:
                self.apply_selection(node, view, False)

    def apply_selection(self, index, view, is_selected):
        '''Applies the selection to the view. This is called internally when
        a view is displayed and it needs to be shown as selected or as not
        selected.

        It is called when :meth:`select_node` or :meth:`deselect_node` is
        called or when a view needs to be refreshed. Its function is purely to
        update the view to reflect the selection state. So the function may be
        called multiple times even if the selection state may not have changed.

        If the view is a instance of
        :class:`~kivy.uix.recycleview.views.RecycleDataViewBehavior`, its
        :meth:`~kivy.uix.recycleview.views.RecycleDataViewBehavior.\
apply_selection` method will be called every time the view needs to refresh
        the selection state. Otherwise, the this method is responsible
        for applying the selection.

        :Parameters:

            `index`: int
                The index of the data item that is associated with the view.
            `view`: widget
                The widget that is the view of this data item.
            `is_selected`: bool
                Whether the item is selected.
        '''
        viewclass = view.__class__
        if viewclass not in _view_base_cache:
            _view_base_cache[viewclass] = isinstance(view,
                                                     RecycleDataViewBehavior)

        if _view_base_cache[viewclass]:
            view.apply_selection(self.recycleview, index, is_selected)

    def refresh_view_layout(self, index, layout, view, viewport):
        super(LayoutSelectionBehavior, self).refresh_view_layout(
            index, layout, view, viewport)
        self.apply_selection(index, view, index in self.selected_nodes)


class RecycleLayoutManagerBehavior(object):
    """A RecycleLayoutManagerBehavior is responsible for positioning views into
    the :attr:`RecycleView.data` within a :class:`RecycleView`. It adds new
    views into the data when it becomes visible to the user, and removes them
    when they leave the visible area.
    """

    viewclass = ObjectProperty(None)
    '''See :attr:`RecyclerView.viewclass`.
    '''
    key_viewclass = StringProperty(None)
    '''See :attr:`RecyclerView.key_viewclass`.
    '''

    recycleview = ObjectProperty(None, allownone=True)

    asked_sizes = None

    def attach_recycleview(self, rv):
        self.recycleview = rv
        if rv:
            fbind = self.fbind
            # can be made more selective update than refresh_from_data which
            # causes a full update. But this likely affects most of the data.
            fbind('viewclass', rv.refresh_from_data)
            fbind('key_viewclass', rv.refresh_from_data)
            fbind('viewclass', rv._dispatch_prop_on_source, 'viewclass')
            fbind('key_viewclass', rv._dispatch_prop_on_source,
                  'key_viewclass')

    def detach_recycleview(self):
        self.clear_layout()
        rv = self.recycleview
        if rv:
            funbind = self.funbind
            funbind('viewclass', rv.refresh_from_data)
            funbind('key_viewclass', rv.refresh_from_data)
            funbind('viewclass', rv._dispatch_prop_on_source, 'viewclass')
            funbind('key_viewclass', rv._dispatch_prop_on_source,
                  'key_viewclass')
        self.recycleview = None

    def compute_sizes_from_data(self, data, flags):
        pass

    def compute_layout(self, data, flags):
        pass

    def compute_visible_views(self, data, viewport):
        '''`viewport` is in coordinates of the layout manager.
        '''
        pass

    def set_visible_views(self, indices, data, viewport):
        '''`viewport` is in coordinates of the layout manager.
        '''
        pass

    def refresh_view_layout(self, index, layout, view, viewport):
        '''`See :meth:`~kivy.uix.recycleview.views.RecycleDataAdapter.\
refresh_view_layout`.
        '''
        self.recycleview.view_adapter.refresh_view_layout(
            index, layout, view, viewport)

    def get_view_index_at(self, pos):
        """Return the view `index` on which position, `pos`, falls.

        `pos` is in coordinates of the layout manager.
        """
        pass

    def remove_views(self):
        rv = self.recycleview
        if rv:
            adapter = rv.view_adapter
            if adapter:
                adapter.make_views_dirty()

    def remove_view(self, view, index):
        rv = self.recycleview
        if rv:
            adapter = rv.view_adapter
            if adapter:
                adapter.make_view_dirty(view, index)

    def clear_layout(self):
        rv = self.recycleview
        if rv:
            adapter = rv.view_adapter
            if adapter:
                adapter.invalidate()

    def goto_view(self, index):
        '''Moves the views so that the view corresponding to `index` is
        visible.
        '''
        pass

    def on_viewclass(self, instance, value):
        # resolve the real class if it was a string.
        if isinstance(value, string_types):
            self.viewclass = getattr(Factory, value)



from kivy.tests.common import GraphicUnitTest
from kivy.input.providers.mouse import (
    MouseMotionEventProvider as Mouse
)


class MultitouchSimulatorTestCase(GraphicUnitTest):
    framecount = 0

    # helper methods
    def correct_y(self, win, y):
        # flip, because the mouse provider uses system's
        # raw one and it's changed to bottom-left origin
        # with Window's system_size[1] for 'mouse_pos'
        return win.height - y

    def mouse_init(self, on_demand=False, disabled=False, scatter=False):
        # prepare MouseMotionEventProvider
        # and widget it interacts with
        from kivy.base import EventLoop
        from kivy.uix.button import Button
        from kivy.uix.scatter import Scatter

        eventloop = EventLoop
        win = eventloop.window
        eventloop.idle()
        wid = Scatter() if scatter else Button()

        if on_demand:
            mode = 'multitouch_on_demand'
        elif disabled:
            mode = 'disable_multitouch'
        else:
            mode = ''
        mouse = Mouse('unittest', mode)
        mouse.is_touch = True

        # defaults from ME, it's missing because we use
        # the provider directly instead of ME
        mouse.scale_for_screen = lambda *_, **__: None
        mouse.grab_exclusive_class = None
        mouse.grab_list = []

        if on_demand:
            self.assertTrue(mouse.multitouch_on_demand)
        return (eventloop, win, mouse, wid)

    def multitouch_dot_touch(self, button, **kwargs):
        # touch -> dot appears -> touch again -> dot disappears
        eventloop, win, mouse, wid = self.mouse_init(**kwargs)

        # register mouse provider
        mouse.start()
        eventloop.add_input_provider(mouse)

        # no mouse touch anywhere
        self.assertEqual(mouse.counter, 0)
        self.assertEqual(mouse.touches, {})

        # right button down, red dot should appear
        win.dispatch(
            'on_mouse_down',
            10, self.correct_y(win, 10),
            'right', {}
        )
        self.assertEqual(mouse.counter, 1)

        if 'on_demand' in kwargs and 'scatter' not in kwargs:
            # doesn't do anything on a pure Button
            self.render(wid)

            # cleanup!
            # remove mouse provider
            mouse.stop()
            eventloop.remove_input_provider(mouse)
            return

        elif 'on_demand' in kwargs and 'scatter' in kwargs:
            self.assertIn(
                'multitouch_sim',
                mouse.touches['mouse1'].profile
            )
            self.assertTrue(mouse.multitouch_on_demand)

            # multitouch_sim is changed in on_touch_down
            # method of the widget that's able to handle
            # multiple touches, therefore for Scatter we
            # need to dispatch the method and because we
            # triggered only on_mouse_down directly i.e.
            # without ME dispatch, on_touch_down was not
            # called == multitouch_sim is False
            self.advance_frames(1)  # initialize stuff
            wid.on_touch_down(mouse.touches['mouse1'])
            wid.on_touch_up(mouse.touches['mouse1'])
            self.assertTrue(mouse.touches['mouse1'].multitouch_sim)

        elif 'disabled' in kwargs:
            self.assertIsNone(
                mouse.touches['mouse1'].ud.get('_drawelement')
            )  # the red dot isn't present

        else:
            self.assertIsNotNone(
                mouse.touches['mouse1'].ud.get('_drawelement')
            )  # the red dot is present

        # XXX right button up
        # first release the touch then check, so that we
        # have the red dot drawn in on_demand and in the
        # default (multitouch everywhere) because in the
        # multitouch_on_demand is the circle drawn after
        # the touch is released (in on_mouse_release)
        win.dispatch(
            'on_mouse_up',
            10, self.correct_y(win, 10),
            'right', {}
        )

        self.assertEqual(mouse.counter, 1)

        # because the red dot is removed by the left button
        if 'disabled' not in kwargs:
            self.assertIn('mouse1', mouse.touches)
            self.assertIsNotNone(
                mouse.touches['mouse1'].ud.get('_drawelement')
            )  # the red dot is present

        # button is down on the previous dot's position
        win.dispatch(
            'on_mouse_down',
            10, self.correct_y(win, 10),
            button, {}
        )
        # if the multitouch is disabled, the touch event
        # increments the counter
        self.assertEqual(
            mouse.counter,
            1 + int('disabled' in kwargs)
        )
        if 'disabled' in kwargs:
            # the right click is ignored, test ends here
            self.assertNotIn(
                'mouse1', mouse.touches
            )
            # cleanup!
            # remove mouse provider
            mouse.stop()
            eventloop.remove_input_provider(mouse)
            return
        else:
            self.assertIsNotNone(
                mouse.touches['mouse1'].ud.get('_drawelement')
            )  # the red dot is present

        # ellipse proxy (<3 #1318 Instruction.proxy_ref)
        dot_proxy = mouse.touches[
            'mouse1'
        ].ud.get('_drawelement')[1].proxy_ref

        # the dot is removed after the touch is released
        # when right - touch is preserved -> dot remains
        # when left  - touch is destroyed -> dot removed
        win.dispatch(
            'on_mouse_up',
            10, self.correct_y(win, 10),
            button, {}
        )  # no matter where

        # the touch, which holds the only ref to the dot
        # instance (Ellipse) is collected, therefore the
        # proxy can confirm the dot is removed
        # (indirect ref at least + it would be nasty for
        # checking if the ellipse remained on visible on
        # the Canvas after being GC-ed if not impossible
        # without the Instruction object trick ._. )
        if button == 'left':
            with self.assertRaises(ReferenceError):
                print(dot_proxy)

            self.assertEqual(mouse.counter, 1)
            self.assertNotIn('mouse1', mouse.touches)
            self.assertEqual(mouse.touches, {})

        elif button == 'right':
            self.assertEqual(mouse.counter, 1)
            self.assertIn('mouse1', mouse.touches)
            self.assertIsNotNone(
                mouse.touches['mouse1'].ud.get('_drawelement')
            )  # the red dot is present

        self.render(wid)

        # cleanup!
        # remove mouse provider
        mouse.stop()
        eventloop.remove_input_provider(mouse)

    def multitouch_dot_move(self, button, **kwargs):
        # touch -> dot appears -> move touch -> dot moves
        # -> release touch -> touch & dot disappear
        eventloop, win, mouse, wid = self.mouse_init(**kwargs)

        # register mouse provider
        mouse.start()
        eventloop.add_input_provider(mouse)

        # no mouse touch anywhere
        self.assertEqual(mouse.counter, 0)
        self.assertEqual(mouse.touches, {})

        # right button down, red dot should appear
        # if the 'multitouch_on_demand' is disabled
        win.dispatch(
            'on_mouse_down',
            10, self.correct_y(win, 10),
            'right', {}
        )
        self.assertEqual(mouse.counter, 1)

        if 'on_demand' in kwargs and 'scatter' not in kwargs:
            # doesn't do anything on a pure Button
            self.render(wid)

            # cleanup!
            # remove mouse provider
            mouse.stop()
            eventloop.remove_input_provider(mouse)
            return

        # XXX right button up
        # first release the touch then check, so that we
        # have the red dot drawn in on_demand and in the
        # default (multitouch everywhere) because in the
        # multitouch_on_demand is the circle drawn after
        # the touch is released (in on_mouse_release)
        elif 'on_demand' in kwargs and 'scatter' in kwargs:
            # on_demand works after the touch is up
            self.assertIn(
                'multitouch_sim',
                mouse.touches['mouse1'].profile
            )
            self.assertTrue(mouse.multitouch_on_demand)

            # multitouch_sim is changed in on_touch_down
            # method of the widget that's able to handle
            # multiple touches, therefore for Scatter we
            # need to dispatch the method and because we
            # triggered only on_mouse_down directly i.e.
            # without ME dispatch, on_touch_down was not
            # called == multitouch_sim is False
            self.advance_frames(1)  # initialize stuff
            wid.on_touch_down(mouse.touches['mouse1'])
            wid.on_touch_up(mouse.touches['mouse1'])
            self.assertTrue(mouse.touches['mouse1'].multitouch_sim)

            win.dispatch(
                'on_mouse_up',
                10, self.correct_y(win, 10),
                'right', {}
            )
            color = mouse.touches[
                'mouse1'
            ].ud.get('_drawelement')[0].proxy_ref
            ellipse = mouse.touches[
                'mouse1'
            ].ud.get('_drawelement')[1].proxy_ref
            win.dispatch(
                'on_mouse_down',
                10, self.correct_y(win, 10),
                'right', {}
            )

        elif 'disabled' in kwargs:
            self.assertIsNone(
                mouse.touches['mouse1'].ud.get('_drawelement')
            )  # the red dot isn't present

        else:
            self.assertIsNotNone(
                mouse.touches['mouse1'].ud.get('_drawelement')
            )  # the red dot is present

        # do NOT make any hard refs to '_drawelement'
        if 'disabled' in kwargs:
            # the right click doesn't draw the red dot
            # the instructions aren't present, test ends
            self.assertIsNone(
                mouse.touches['mouse1'].ud.get('_drawelement')
            )  # the red dot isn't present
            # cleanup!
            # remove mouse provider
            mouse.stop()
            eventloop.remove_input_provider(mouse)
            return

        else:
            color = mouse.touches[
                'mouse1'
            ].ud.get('_drawelement')[0].proxy_ref
            ellipse = mouse.touches[
                'mouse1'
            ].ud.get('_drawelement')[1].proxy_ref

        # the red dot moves when the touch is moving
        win.dispatch(
            'on_mouse_move',
            11, self.correct_y(win, 11),
            {}
        )
        self.assertEqual(
            ellipse.pos,
            (1, self.correct_y(win, win.height - 1))
        )  # bounding box from Rectangle, R=10 -> 20 width

        # right button up
        win.dispatch(
            'on_mouse_up',
            10, self.correct_y(win, 10),
            'right', {}
        )
        self.assertEqual(mouse.counter, 1)

        # because the red dot is removed by the left button
        self.assertIn('mouse1', mouse.touches)
        self.assertIsNotNone(
            mouse.touches['mouse1'].ud.get('_drawelement')
        )  # the red dot is present

        # the dot is at (11, 11), but the touch is in
        # its bounding box, therefore it can move it
        win.dispatch(
            'on_mouse_down',
            10, self.correct_y(win, 10),
            button, {}
        )

        # manipulating already existing touch,
        # no new one was created
        self.assertEqual(mouse.counter, 1)
        self.assertIsNotNone(
            mouse.touches['mouse1'].ud.get('_drawelement')
        )  # the red dot is present

        # the red dot moves when the touch is moving
        win.dispatch(
            'on_mouse_move',
            50, self.correct_y(win, 50),
            {}
        )
        self.assertEqual(
            ellipse.pos,
            (40, self.correct_y(win, win.height - 40))
        )  # bounding box from Rectangle, R=10 -> 20 width

        # the dot is removed after the touch is released
        # when right - touch is preserved -> dot remains
        # when left  - touch is destroyed -> dot removed
        win.dispatch(
            'on_mouse_up',
            10, self.correct_y(win, 10),
            button, {}
        )  # no matter where
        self.assertEqual(mouse.counter, 1)

        if button == 'left':
            self.assertNotIn('mouse1', mouse.touches)
        elif button == 'right':
            self.assertIn('mouse1', mouse.touches)
            self.assertIsNotNone(
                mouse.touches['mouse1'].ud.get('_drawelement')
            )  # the red dot is present

        self.render(wid)

        # cleanup!
        # remove mouse provider
        mouse.stop()
        eventloop.remove_input_provider(mouse)

    # tests
    def test_multitouch_dontappear(self):
        eventloop, win, mouse, wid = self.mouse_init()

        # register mouse provider
        mouse.start()
        eventloop.add_input_provider(mouse)

        # no mouse touch anywhere
        self.assertEqual(mouse.counter, 0)
        self.assertEqual(mouse.touches, {})

        # left button down
        win.dispatch(
            'on_mouse_down',
            10, self.correct_y(win, 10),
            'left', {}
        )
        win.dispatch(
            'on_mouse_move',
            11, self.correct_y(win, 11),
            {}
        )
        self.assertEqual(mouse.counter, 1)
        self.assertIsNone(
            mouse.touches['mouse1'].ud.get('_drawelement')
        )  # the red dot isn't present

        # left button up
        win.dispatch(
            'on_mouse_up',
            10, self.correct_y(win, 10),
            'left', {}
        )
        # after the releasing the touch disappears,
        # but the counter remains
        self.assertEqual(mouse.counter, 1)
        self.assertNotIn('mouse1', mouse.touches)

        self.advance_frames(1)
        self.render(wid)

        # cleanup!
        # remove mouse provider
        mouse.stop()
        eventloop.remove_input_provider(mouse)

    def test_multitouch_appear(self):
        eventloop, win, mouse, wid = self.mouse_init()

        # register mouse provider
        mouse.start()
        eventloop.add_input_provider(mouse)

        # no mouse touch anywhere
        self.assertEqual(mouse.counter, 0)
        self.assertEqual(mouse.touches, {})

        # right button down, red dot should appear
        win.dispatch(
            'on_mouse_down',
            10, self.correct_y(win, 10),
            'right', {}
        )
        self.assertEqual(mouse.counter, 1)
        self.assertIsNotNone(
            mouse.touches['mouse1'].ud.get('_drawelement')
        )  # the red dot is present

        # do NOT make any hard refs to '_drawelement'
        color = mouse.touches[
            'mouse1'
        ].ud.get('_drawelement')[0].proxy_ref
        ellipse = mouse.touches[
            'mouse1'
        ].ud.get('_drawelement')[1].proxy_ref

        # check ellipse's position
        self.assertEqual(
            ellipse.pos[0], 0
        )  # bounding box from Rectangle, R=10 -> 20 width
        # almost equal because the correct_y uses the same
        # float - float, which returns decimal garbage
        self.assertAlmostEqual(
            ellipse.pos[1],
            self.correct_y(win, win.height),
            delta=0.0001
        )

        win.dispatch(
            'on_mouse_move',
            11, self.correct_y(win, 11),
            {}
        )
        # the red dot moves when the touch is moving
        self.assertEqual(
            ellipse.pos,
            (1, self.correct_y(win, win.height - 1))
        )  # bounding box from Rectangle, R=10 -> 20 width
        win.dispatch(
            'on_mouse_up',
            10, self.correct_y(win, 10),
            'right', {}
        )

        self.assertEqual(
            ellipse.pos,
            (1, self.correct_y(win, win.height - 1))
        )  # bounding box from Rectangle, R=10 -> 20 width
        self.assertEqual(mouse.counter, 1)
        # because the red dot is removed by the left button
        self.assertIn('mouse1', mouse.touches)
        self.assertIsNotNone(
            mouse.touches['mouse1'].ud.get('_drawelement')
        )  # the red dot is present

        self.render(wid)

        # cleanup!
        # remove mouse provider
        mouse.stop()
        eventloop.remove_input_provider(mouse)

    def test_multitouch_dot_lefttouch(self):
        self.multitouch_dot_touch('left')

    def test_multitouch_dot_leftmove(self):
        self.multitouch_dot_move('left')

    def test_multitouch_dot_righttouch(self):
        self.multitouch_dot_touch('right')

    def test_multitouch_dot_rightmove(self):
        self.multitouch_dot_move('right')

    def test_multitouch_on_demand_noscatter_lefttouch(self):
        self.multitouch_dot_touch('left', on_demand=True)

    def test_multitouch_on_demand_noscatter_leftmove(self):
        self.multitouch_dot_move('left', on_demand=True)

    def test_multitouch_on_demand_noscatter_righttouch(self):
        self.multitouch_dot_touch('right', on_demand=True)

    def test_multitouch_on_demand_noscatter_rightmove(self):
        self.multitouch_dot_move('right', on_demand=True)

    def test_multitouch_on_demand_scatter_lefttouch(self):
        self.multitouch_dot_touch(
            'left', on_demand=True, scatter=True
        )

    def test_multitouch_on_demand_scatter_leftmove(self):
        self.multitouch_dot_move(
            'left', on_demand=True, scatter=True
        )

    def test_multitouch_on_demand_scatter_righttouch(self):
        self.multitouch_dot_touch(
            'right', on_demand=True, scatter=True
        )

    def test_multitouch_on_demand_scatter_rightmove(self):
        self.multitouch_dot_move(
            'right', on_demand=True, scatter=True
        )

    def test_multitouch_disabled_lefttouch(self):
        self.multitouch_dot_touch('left', disabled=True)

    def test_multitouch_disabled_leftmove(self):
        self.multitouch_dot_move('left', disabled=True)

    def test_multitouch_disabled_righttouch(self):
        self.multitouch_dot_touch('right', disabled=True)

    def test_multitouch_disabled_rightmove(self):
        self.multitouch_dot_move('right', disabled=True)


if __name__ == '__main__':
    import unittest
    unittest.main()



'''
Bubble
======

Test of the widget Bubble.
'''

from kivy.app import App
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.button import Button
from kivy.lang import Builder
from kivy.uix.bubble import Bubble

Builder.load_string('''
<cut_copy_paste>
    size_hint: (None, None)
    size: (160, 120)
    pos_hint: {'center_x': .5, 'y': .6}
    BubbleButton:
        text: 'Cut'
    BubbleButton:
        text: 'Copy'
    BubbleButton:
        text: 'Paste'
''')


class cut_copy_paste(Bubble):
    pass


class BubbleShowcase(FloatLayout):

    def __init__(self, **kwargs):
        super(BubbleShowcase, self).__init__(**kwargs)
        self.but_bubble = Button(text='Press to show bubble')
        self.but_bubble.bind(on_release=self.show_bubble)
        self.add_widget(self.but_bubble)

    def show_bubble(self, *l):
        if not hasattr(self, 'bubb'):
            self.bubb = bubb = cut_copy_paste()
            self.add_widget(bubb)
        else:
            values = ('left_top', 'left_mid', 'left_bottom', 'top_left',
                'top_mid', 'top_right', 'right_top', 'right_mid',
                'right_bottom', 'bottom_left', 'bottom_mid', 'bottom_right')
            index = values.index(self.bubb.arrow_pos)
            self.bubb.arrow_pos = values[(index + 1) % len(values)]


class TestBubbleApp(App):

    def build(self):
        return BubbleShowcase()


if __name__ == '__main__':
    TestBubbleApp().run()
